from enum import Enum

from logger_local.LoggerComponentEnum import LoggerComponentEnum


# TODO Move everything related to sync to separate directory called
#  sync_data_source (preferable with it's own src and tests directories)
class UpdateStatus(Enum):
    UPDATE_DATA_SOURCE = -1
    DONT_UPDATE = 0
    UPDATE_CIRCLEZ = 1  # TODO Don't use CIRCLEZ


class IsTestData(Enum):
    PRODUCTION = 0
    TEST_DATA = 1
    TEMP_DATA = 2


# connector / cursor
DATABASE_MYSQL_PYTHON_PACKAGE_COMPONENT_ID = 112
DATABASE_MYSQL_PYTHON_PACKAGE_COMPONENT_NAME = 'database_mysql_local\\connector'
CONNECTOR_DEVELOPER_EMAIL = 'idan.a@circ.zone'
LOGGER_CONNECTOR_CODE_OBJECT = {
    'component_id': DATABASE_MYSQL_PYTHON_PACKAGE_COMPONENT_ID,
    'component_name': DATABASE_MYSQL_PYTHON_PACKAGE_COMPONENT_NAME,
    'component_category': LoggerComponentEnum.ComponentCategory.Code.value,
    'developer_email_address': CONNECTOR_DEVELOPER_EMAIL
}

# generic_crud
DATABASE_MYSQL_PYTHON_GENERIC_CRUD_COMPONENT_ID = 206
DATABASE_MYSQL_PYTHON_GENERIC_CRUD_COMPONENT_NAME = 'database_mysql_local\\generic_crud'
GENERIC_CRUD_DEVELOPER_EMAIL = 'akiva.s@circ.zone'
CRUD_MYSQL_CODE_LOGGER_OBJECT = {
    'component_id': DATABASE_MYSQL_PYTHON_GENERIC_CRUD_COMPONENT_ID,
    'component_name': DATABASE_MYSQL_PYTHON_GENERIC_CRUD_COMPONENT_NAME,
    'component_category': LoggerComponentEnum.ComponentCategory.Code.value,
    'developer_email_address': GENERIC_CRUD_DEVELOPER_EMAIL
}

# generic_crud_ml
DATABASE_MYSQL_GENERIC_CRUD_ML_COMPONENT_ID = 7001
DATABASE_MYSQL_GENERIC_CRUD_ML_COMPONENT_NAME = 'database_mysql_local\\generic_crud_ml'
GENERIC_CRUD_ML_DEVELOPER_EMAIL = 'tal.g@circ.zone'
CRUD_ML_CODE_LOGGER_OBJECT = {
    'component_id': DATABASE_MYSQL_GENERIC_CRUD_ML_COMPONENT_ID,
    'component_name': DATABASE_MYSQL_GENERIC_CRUD_ML_COMPONENT_NAME,
    'component_category': LoggerComponentEnum.ComponentCategory.Code.value,
    'developer_email_address': GENERIC_CRUD_ML_DEVELOPER_EMAIL
}

# generic_mapping
DATABASE_MYSQL_PYTHON_GENERIC_MAPPING_COMPONENT_ID = 7002
DATABASE_MYSQL_PYTHON_GENERIC_MAPPING_COMPONENT_NAME = 'database_mysql_local\\generic_mapping'
GENERIC_MAPPING_DEVELOPER_EMAIL = 'sahar.g@circ.zone'
MAPPING_CODE_LOGGER_OBJECT = {
    'component_id': DATABASE_MYSQL_PYTHON_GENERIC_MAPPING_COMPONENT_ID,
    'component_name': DATABASE_MYSQL_PYTHON_GENERIC_MAPPING_COMPONENT_NAME,
    'component_category': LoggerComponentEnum.ComponentCategory.Code.value,
    'developer_email_address': GENERIC_MAPPING_DEVELOPER_EMAIL
}



# TODO Move everything related to sync to separate directory called sync_data_source (preferable with it's own src and tests directories)
from datetime import datetime

from .constants_src import UpdateStatus
from .generic_crud import GenericCRUD

# TODO Which packages are using it google-contact? contact-csv?
# TODO: use meta logger


class SyncConflictResolution(GenericCRUD):
    def __init__(self, default_schema_name: str = None, default_table_name: str = None,
                 default_view_table_name: str = None,
                 default_view_with_deleted_and_test_data: str = None,
                 default_column_name: str = None,
                 default_select_clause_value: str = "updated_timestamp",
                 default_where: str = None):
        super().__init__(default_schema_name=default_schema_name, default_table_name=default_table_name,
                         default_view_table_name=default_view_table_name,
                         default_view_with_deleted_and_test_data=default_view_with_deleted_and_test_data,
                         default_column_name=default_column_name,
                         default_select_clause_value=default_select_clause_value, default_where=default_where)

    def get_update_status_by_where(self, *, schema_name: str = None, view_table_name: str = None,
                                   where: str = None, params: tuple, select_clause_value: str = None,
                                   remote_last_modified_timestamp: str) -> UpdateStatus:
        """
        Get the status of the update by the where clause
        :param remote_last_modified_timestamp str
        :param params tuple
        :param schema_name str
        :param view_table_name str
        :param where str
        :param select_clause_value str
        :return str
        """
        schema_name = schema_name or self.default_schema_name
        view_table_name = view_table_name or self.default_view_table_name
        select_clause_value = select_clause_value or self.default_select_clause_value
        where = where or self.default_where
        if not view_table_name or not where or not schema_name:
            self.logger.error("view_table_name, where or schema was not provided")
            return "error"  # TODO: why not raise?
        local_updated_timestamp = self.select_one_value_by_where(
            schema_name=schema_name, view_table_name=view_table_name, select_clause_value=select_clause_value,
            where=where, params=params)
        remote_last_modified_timestamp: datetime = datetime.strptime(remote_last_modified_timestamp,
                                                                     '%Y-%m-%d %H:%M:%S')
        if local_updated_timestamp is None or remote_last_modified_timestamp > local_updated_timestamp:
            return UpdateStatus.UPDATE_CIRCLEZ
        elif remote_last_modified_timestamp < local_updated_timestamp:
            return UpdateStatus.UPDATE_DATA_SOURCE
        else:
            return UpdateStatus.DONT_UPDATE

    def get_update_status_by_id(self, *, schema_name: str = None, view_table_name: str = None,
                                column_name: str = None, column_value: str = None,
                                remote_last_modified_timestamp: str,
                                select_clause_value: str = None) -> UpdateStatus:
        """
        Get the status of the update by the id
        :param remote_last_modified_timestamp: str
        :param schema_name str
        :param view_table_name str
        :param column_name str
        :param column_value str
        :param select_clause_value str
        :return str
        """
        schema_name = schema_name or self.default_schema_name
        view_table_name = view_table_name or self.default_view_table_name
        column_name = column_name or self.default_column_name
        if not view_table_name or not column_name or not schema_name:
            self.logger.error("view_table_name, column_name or schema was not provided")
            return "error"  # TODO: why not raise?
        local_updated_timestamp = self.select_one_value_by_column_and_value(
            schema_name=schema_name, view_table_name=view_table_name, select_clause_value=select_clause_value,
            column_name=column_name, column_value=column_value)
        remote_last_modified_timestamp: datetime = datetime.strptime(remote_last_modified_timestamp,
                                                                     '%Y-%m-%d %H:%M:%S')
        if local_updated_timestamp is None or remote_last_modified_timestamp > local_updated_timestamp:
            return UpdateStatus.UPDATE_CIRCLEZ
        elif remote_last_modified_timestamp < local_updated_timestamp:
            return UpdateStatus.UPDATE_DATA_SOURCE
        else:
            return UpdateStatus.DONT_UPDATE

    # TODO: complete this
    def get_update_status_and_information_by_where(self, *, schema_name: str = None, view_table_name: str = None,
                                                   where: str = None, params: tuple, select_clause_value: str = "*",
                                                   remote_last_modified_timestamp: str,
                                                   local_last_modified_column_name: str) -> dict:
        """
        Get the status of the update by the where clause
        :param remote_last_modified_timestamp str
        :param params tuple
        :param schema_name str
        :param view_table_name str
        :param where str
        :param select_clause_value str
        :return dict
        """
        schema_name = schema_name or self.default_schema_name
        view_table_name = view_table_name or self.default_view_table_name
        select_clause_value = select_clause_value or self.default_select_clause_value
        where = where or self.default_where
        if not view_table_name or not where or not schema_name:
            self.logger.error("view_table_name, where or schema was not provided")
            raise ValueError("view_table_name, where or schema was not provided")
        row_dict = self.select_one_dict_by_where(
            schema_name=schema_name, view_table_name=view_table_name, select_clause_value=select_clause_value,
            where=where, params=params)
        result_dict = row_dict
        local_updated_timestamp = row_dict.get(local_last_modified_column_name)
        remote_last_modified_timestamp: datetime = datetime.strptime(remote_last_modified_timestamp,
                                                                     '%Y-%m-%d %H:%M:%S')
        if local_updated_timestamp is None or remote_last_modified_timestamp > local_updated_timestamp:
            result_dict["update_status"] = UpdateStatus.UPDATE_CIRCLEZ
        elif remote_last_modified_timestamp < local_updated_timestamp:
            result_dict["update_status"] = UpdateStatus.UPDATE_DATA_SOURCE
        else:
            result_dict["update_status"] = UpdateStatus.DONT_UPDATE
        return result_dict

    def get_update_status_and_information_list_by_where(self, *, schema_name: str = None,
                                                        view_table_name: str = None, where: str = None,
                                                        params: tuple, select_clause_value: str = "*",
                                                        remote_last_modified_timestamp: str,
                                                        local_last_modified_column_name: str) -> list[dict]:
        """
        Get the status of the update by the where clause
        :param remote_last_modified_timestamp str
        :param params tuple
        :param schema_name str
        :param view_table_name str
        :param where str
        :param select_clause_value str
        :return dict
        """
        schema_name = schema_name or self.default_schema_name
        view_table_name = view_table_name or self.default_view_table_name
        select_clause_value = select_clause_value or self.default_select_clause_value
        remote_last_modified_timestamp: datetime = datetime.strptime(remote_last_modified_timestamp,
                                                                     '%Y-%m-%d %H:%M:%S')
        where = where or self.default_where
        if not view_table_name or not where or not schema_name:
            self.logger.error("view_table_name, where or schema was not provided")
            raise ValueError("view_table_name, where or schema was not provided")
        row_dict_list: list[dict] = self.select_multi_dict_by_where(
            schema_name=schema_name, view_table_name=view_table_name, select_clause_value=select_clause_value,
            where=where, params=params)
        result_dict_list: list[dict] = row_dict_list
        for result_dict in result_dict_list:
            local_updated_timestamp = result_dict.get(local_last_modified_column_name)
            if local_updated_timestamp is None or remote_last_modified_timestamp > local_updated_timestamp:
                result_dict["update_status"] = UpdateStatus.UPDATE_CIRCLEZ
            elif remote_last_modified_timestamp < local_updated_timestamp:
                result_dict["update_status"] = UpdateStatus.UPDATE_DATA_SOURCE
            else:
                result_dict["update_status"] = UpdateStatus.DONT_UPDATE
        return result_dict_list



# PACKAGE_VERSION from setup.py generated by MakePy.ps1
PACKAGE_VERSION = '0.1.51'


# TODO Move get_package_version() to python-sdk (generated by MakePy.ps1)
# TODO Shall we make it method in all of our classes?
def get_package_version() -> str:
    return PACKAGE_VERSION



import copy
import inspect
import os
from functools import lru_cache
from typing import Any, Optional, Tuple

from logger_local.LoggerLocal import Logger
from python_sdk_remote.utilities import get_environment_name
from python_sdk_remote.utilities import our_get_env
from url_remote.environment_name_enum import EnvironmentName

from .constants_src import LOGGER_CONNECTOR_CODE_OBJECT
from .table_columns import table_columns
from .table_definition import table_definition
# from .to_sql_interface import ToSQLInterface
from database_infrastructure_local.to_sql_interface import ToSQLInterface

logger = Logger.create_logger(object=LOGGER_CONNECTOR_CODE_OBJECT)


def get_sql_hostname() -> str:
    sql_hostname = our_get_env("RDS_HOSTNAME")
    sql_hostname = sql_hostname.strip('"\'')
    return sql_hostname


def get_sql_username() -> str:
    sql_username = our_get_env("RDS_USERNAME")
    return sql_username


def get_sql_password() -> str:
    sql_password = our_get_env("RDS_PASSWORD")
    return sql_password


def get_sql_port() -> str:
    sql_port = our_get_env("RDS_PORT", default="3306")
    return sql_port


def get_ssh_hostname() -> str:
    return our_get_env("SSH_HOSTNAME", raise_if_not_found=False)


def get_ssh_username() -> str:
    ssh_username = our_get_env("SSH_USERNAME", raise_if_empty=True)
    return ssh_username


def get_ssh_port() -> str:
    ssh_port = our_get_env("SSH_PORT", raise_if_empty=True)
    return ssh_port


def get_private_key_file_path() -> str:
    private_key_file_path = our_get_env("PRIVATE_KEY_FILE_PATH", raise_if_empty=True)
    return private_key_file_path


# TODO: class DatabaseMysqlUtils(metaclass=MetaLogger, object=object_database):
def validate_select_table_name(view_table_name: str, is_ignore_duplicate: bool = False) -> None:
    # TODO: try to detect the table name from the view name (with warning)
    if (get_environment_name() not in (EnvironmentName.DVLP1.value, EnvironmentName.PROD1.value)
            and not view_table_name.endswith("_view") and not is_ignore_duplicate):
        raise Exception(
            f"View name must end with '_view' in this environment (got {view_table_name})")


def validate_none_select_table_name(table_name: str) -> None:
    if (get_environment_name() not in (EnvironmentName.DVLP1.value, EnvironmentName.PROD1.value)
            and not table_name.endswith("_table")):
        raise Exception(
            f"Table name must end with '_table' in this environment  (got {table_name})")


# TODO Shall we change it from a function to a method in GenericCrud
def process_insert_data_dict(data_dict: dict | None) -> Tuple[str, str, tuple]:
    """Example:
    Input: {"name": "John", "coordinate": Point(1, 2)}
    Output: ("`name`, `coordinate`",
             "%s, Point(1, 2)",
            ("John", ))
    """
    if not data_dict:
        columns_str, values_str, params = '', '', ()
        return columns_str, values_str, params

    columns = []
    values = []
    params = tuple()

    for key, value in data_dict.items():
        # new_key = f'`{key}`' if '.' in key else key
        if key.startswith("`") and key.endswith("`"):
            new_key = key
        else:
            new_key = f"`{key}`"

        columns.append(new_key)

        is_instance = isinstance(value, ToSQLInterface)

        if is_instance:
            values.append(value.to_sql())
        else:
            values.append('%s')
            params += (value,)

    columns_str = ','.join(columns)
    values_str = ','.join(values)

    return columns_str, values_str, params


def process_upsert_data_dict(data_dict: dict | None, compare_with_or: bool,
                             where_compare: str | None = None,
                             params_compare: tuple | None = None) -> Tuple[str, tuple]:

    where_clauses = []
    params = []
    for column, value in data_dict.items():
        if isinstance(value, list):
            where_clauses.append(f"({' OR '.join([f'{column}=%s' for _ in value])})")
            params.extend(value)
        elif isinstance(value, ToSQLInterface):
            where_clauses.append(f"{column}={value.to_sql()}")
        elif value is None:
            where_clauses.append(f"{column} IS NULL")
        else:
            where_clauses.append(f"{column}=%s")
            params.append(value)

    where_clause = " OR " if compare_with_or else " AND "
    where_clause = where_clause.join(where_clauses)
    if where_compare and params_compare:
        if where_clause:
            where_clause = f"({where_clause}) AND ({where_compare})"
        else:
            where_clause = where_compare
        params += params_compare
    return where_clause, params


def process_update_data_dict(data_dict: dict | None) -> Tuple[str, tuple]:
    """Example:
    Input: {"name": "John", "coordinate": Point(1, 2)}
    Output: ("name=%s, coordinate=Point(1, 2)", ("John", ))
    """
    if not data_dict:
        set_values_str, params = '', {}
        return set_values_str, params

    set_values = []
    params = tuple()
    for key, value in data_dict.items():
        if isinstance(value, ToSQLInterface):
            set_values.append(f"`{key}`={value.to_sql()}")
        else:
            set_values.append(f"`{key}`=%s")
            params += (value,)

    # + "," because we add updated_timestamp in the update query
    set_values_str = ', '.join(set_values) + ","
    return set_values_str, params


def process_select_data_dict(data_dict: dict | None, select_with_or: bool = False) -> Tuple[str, tuple]:
    """Example:
    Input: {"name": "John", "coordinate": Point(1, 2)}
    Output: ("name=%s AND coordinate=Point(1, 2)", ("John", ))
    """
    if not data_dict:
        where_clause, params = '', {}
        return where_clause, params

    where_clauses = []
    params = tuple()
    for key, value in data_dict.items():
        if isinstance(value, list):
            where_clauses.append(f"({' OR '.join([f'{key}=%s' for _ in value])})")
            params += tuple(value)
        elif isinstance(value, ToSQLInterface):
            where_clauses.append(f"{key}={value.to_sql()}")
        elif value is None:
            where_clauses.append(f"{key} IS NULL")
        else:
            where_clauses.append(f"{key}=%s")
            params += (value,)

    where_clause = " OR " if select_with_or else " AND "
    where_clause = where_clause.join(where_clauses)
    return where_clause, params


@lru_cache
def detect_if_is_test_data() -> bool:
    """Check if running from a Unit Test file."""
    possible_current_files = [os.path.basename(frame.filename) for frame in inspect.stack()]
    is_test_data = any(file_name.startswith('test_') or file_name.endswith('_test.py') or "pytest" in file_name
                       for file_name in possible_current_files)
    return is_test_data


def get_entity_type_ids_by_table_name(table_name: str) -> int | None:
    """Get the entity_type_id1 and entity_type_id2 from the table_definition by table_name **IF** the table is many-to-many.

    :param table_name: The name of the table.
    :return: A tuple of entity_type_id1 and entity_type_id2 if the table is many-to-many, otherwise None.
    """
    entity_type_id1 = get_entity_type_id1_by_table_name(table_name)
    entity_type_id2 = get_entity_type_id2_by_table_name(table_name)

    if entity_type_id1 is not None and entity_type_id2 is not None:
        return entity_type_id1, entity_type_id2
    else:
        logger.warning(f"Table {table_name} is not a many-to-many table, "
                       f"entity_type_id1: {entity_type_id1}, entity_type_id2: {entity_type_id2}. ")
        return None


def get_entity_type_id1_by_table_name(table_name: str) -> int | None:
    """
    get the entity_type_id1 from the table_definition by table_name.

    :param table_name: The name of the table.
    :return: The entity_type_id1 of the table, or None if not found.
    """
    # TODO Should we search for it in entity_type_table or table_definition?

    entity_type_table = table_definition.get(table_name, {})
    entity_type_id1 = entity_type_table.get("entity_type_id1")

    return entity_type_id1


def get_entity_type_id2_by_table_name(table_name: str) -> int | None:
    """
    get the entity_type_id2 from the table_definition by table_name.

    :param table_name: The name of the table.
    :return: The entity_type_id2 of the table, or None if not found.
    """
    entity_type_table = table_definition.get(table_name, {})

    entity_type_id2 = entity_type_table.get("entity_type_id2")
    return entity_type_id2


# def generate_table_name(entity_name: Optional[str],
#                         schema_name: Optional[str]) -> Optional[str]:
#     if entity_name:
#         table_name = entity_name + "_table"
#         return table_name
#     if schema_name:
#         table_name = schema_name + "_table"
#         return table_name


def generate_table_name(entity_name: str = None,
                        schema_name: str = None,
                        ml_table_name: str = None) -> str:
    if ml_table_name:
        table_name = (ml_table_name or "").replace("_ml", "")
    elif entity_name:
        table_name = entity_name + "_table"
    elif schema_name:
        table_name = schema_name + "_table"
    else:
        table_name = None

    # Append env suffix if needed
    table_name_with_env = add_env_to_suffix(table_name)

    return table_name_with_env


def generate_view_name(table_name: Optional[str]) -> Optional[str]:
    if table_name:
        view_name = table_name.replace("_table", "_view")
    else:
        view_name = table_name

    # Append env suffix if needed
    view_name_with_env = add_env_to_suffix(view_name)

    return view_name_with_env


def add_env_to_suffix(name: Optional[str]) -> Optional[str]:
    """
    Add environment name to the suffix of the name and update schema name
    for specific tables/views in multi-environment databases.

    Examples (env = 'play1'):
    - profile_table -> test_play1.profile_play1_table
    - user_view     -> test_play1.user_play1_view

    All other names are returned unchanged.
    """
    if name is None:
        return None

    env_name = get_environment_name()
    if not env_name:
        return name

    # Multi-environment entities that require schema + env
    special_entities = {
        "test_county_table", "test_county_view",
        "profile_table", "profile_view",
        "contact_table", "contact_view",
        "user_table", "user_view",
        "person_table", "person_view",
        "user-external_table", "user-external_view"
    }

    if name in special_entities:
        for suffix in ["_table", "_view"]:
            if name.endswith(suffix):
                base = name[: -len(suffix)]
                if not name.endswith(f"_{env_name}{suffix}"):
                    return f"{base}_{env_name}{suffix}"

    return name


def generate_id_column_name(table_name: Optional[str]) -> Optional[str]:
    if table_name:
        # TODO Why play1_ with _ and dvlp1 without _?
        if "_play1_" in table_name or "_dvlp1" in table_name:
            # TODO Why we have environment_name in the table_name in the first place? We should not add the environment_name to the table_name  # noqa E501
            table_name = table_name.replace(f"_{get_environment_name()}_", "_")
        column_name = table_name.replace("_table", "_id")
        return column_name


def validate_single_clause_value(select_clause_value: str = None) -> None:
    if not select_clause_value or "," in select_clause_value or select_clause_value == "*":
        raise ValueError(f"select value requires a single column name, got {select_clause_value}")


def get_where_params(column_name: str, column_value: Any) -> tuple:
    # If we use "if column_value:" it will not work for 0, False, etc.
    if not column_name:
        raise ValueError(f"column_name is required, got {column_name}")
    if isinstance(column_value, ToSQLInterface):
        where = f"`{column_name}`={column_value.to_sql()}"
        params = None
    elif column_value is not None:
        if isinstance(column_value, (list, tuple, set)):
            where = f"`{column_name}` IN (" + ",".join(["%s"] * len(column_value)) + ")"
            params = tuple(column_value)
        else:
            where = f"`{column_name}`=%s"
            params = (column_value,)
    else:
        where = f"`{column_name}` IS NULL"
        params = None
    return where, params


def where_skip_null_values(where: str | None, select_clause_value: str,
                           skip_null_values: bool = True) -> str:
    if skip_null_values:
        validate_single_clause_value(select_clause_value)
        where_skip = f"`{select_clause_value}` IS NOT NULL"
        if where:
            where += f" AND {where_skip}"
        else:
            where = where_skip
    return where


@lru_cache(maxsize=64)
def replace_view_with_table(view_table_name: str, select_clause_value: str = None) -> str:
    # test data does not appear in the view, but we still wants to access it in tests.
    if not view_table_name:
        return view_table_name
    # Guess the table name from the view name:
    table_name = remove_tag(view_table_name.replace("_view", "_table"))
    scan_table_definition_for_table_name = table_definition.get(table_name, {}).get("view_name") != view_table_name
    if scan_table_definition_for_table_name:
        for table, values in table_definition.items():
            if values["view_name"] == view_table_name:
                table_name = table  # got a better guess
                break
    if select_clause_value and select_clause_value != "*":
        required_columns = tuple(col.strip() for col in select_clause_value.split(","))  # if columns are specified
    else:
        required_columns = table_columns.get(view_table_name, [])  # all columns in the view

    # Replace 'ST_X(coordinate)' and 'ST_Y(coordinate)' with 'coordinate'
    required_columns = tuple('coordinate' if col in ('ST_X(coordinate)', 'ST_Y(coordinate)') else
                             remove_tag(col) for col in required_columns)

    # TODO if (is_debug) # Less optimized for runtime, better for debug
    if table_name in table_columns:
        all_columns_present = True
        for col in required_columns:
            if remove_tag(col) not in table_columns.get(table_name, []):
                all_columns_present = False
                logger.warning(f"Column {col} not found in table {table_name}. check the columns in select clause value if they exist in the table.")  # noqa E501
                break

        if all_columns_present:
            view_table_name = table_name

    # if !is_debug # More optimized for runtime, less for debug
    # if table_name in table_columns and all(  # if all required columns from the view present in the table.  # noqa E501
    #         remove_tag(col) in table_columns.get(table_name, []) for col in required_columns):  # noqa E501
    #     view_table_name = table_name

    return view_table_name  # appropriate table not found


def remove_tag(name: str) -> str:
    name_without_tag = name.replace("`", "")
    return name_without_tag


def insert_is_undelete(table_name: str) -> bool:
    is_undelete = table_definition.get(table_name, {}).get("insert_is_undelete")
    return is_undelete


def is_end_timestamp_in_table(table_name: str) -> bool:
    # table_definition outdated for now.
    # is_end_timestamp_in_table = table_definition.get(table_name, {}).get("end_timestamp")
    is_end_timestamp_in_table = is_column_in_table(table_name, "end_timestamp")
    return is_end_timestamp_in_table


def is_column_in_table(table_name: str, column_name: str) -> bool:
    columns = table_columns.get(table_name, [])
    column_in_table = remove_tag(column_name) in columns
    return column_in_table


def get_table_columns(table_name: str = None) -> tuple:
    table_columns_tuple = table_columns.get(table_name, [])
    return table_columns_tuple


def group_list_by_columns(list_of_dicts: list,
                          group_by: str,
                          condense: bool = True) -> dict[tuple | str, list]:
    """if condense is True and there is only one column left in the dict, return the value instead of the dict.
    Examples:
    Input: [{"name": "John", "age": 25}, {"name": "John", "age": 26}], group_by="name"
            -> {"John": [25, 26]}
    Input: [{"name": "John", "age": 25}, {"name": "John", "age": 26}], group_by="name", condense=False
            -> {"John": [{"age": 25}, {"age": 26}]}
    Input: [{"a": 1, "b": 2, "c": 3}, {"a": 1, "b": 2, "c": 4}, {"a": 2, "b": 2, "c": 5}], group_by="a"
            -> {1: [{"b": 2, "c": 3}, {"b": 2, "c": 4}], 2: [{"b": 2, "c": 5}]}
    Input: [{"a": 1, "b": 2, "c": 3}, {"a": 1, "b": 2, "c": 4}, {"a": 2, "b": 2, "c": 5}], group_by="a,b", condense=False
            -> {(1, 2): [{"c": 3}, {"c": 4}], (2, 2): [{"c": 5}]}
    """
    group_by = tuple(map(str.strip, group_by.split(",")))
    grouped = {}
    if not list_of_dicts:
        return grouped
    list_of_dicts = copy.deepcopy(list_of_dicts)
    if any(col not in list_of_dicts[0] for col in group_by):
        raise ValueError(f"{group_by} should be a subset of {tuple(list_of_dicts[0].keys())}")
    if len(group_by) == len(list_of_dicts[0]):
        raise ValueError("Column names in group_by must be less than the number of columns in the list of dicts")

    for dict_row in list_of_dicts:
        key = tuple(dict_row[col] for col in group_by) if len(group_by) > 1 else dict_row[group_by[0]]
        if key not in grouped:
            grouped[key] = []
        for col in group_by:
            dict_row.pop(col)
        grouped[key].append(dict_row if not condense or len(dict_row) > 1 else next(iter(dict_row.values())))
    return grouped


# Not used anymore:
# def extract_duplicated_from_error(error: Exception) -> (Any, str):
# """Error examples:
# - Duplicate entry '1' for key 'test_mysql_table.PRIMARY'  - in such case return the duplicated value, otherwise the column
# - Duplicate entry '7263200721327371865' for key 'test_mysql_table.number_UNIQUE
# - IntegrityError(1062, "1062 (23000): Duplicate entry '202-405-3018' for key 'person_table.person.main_full_number_normalized_UNIQUE'", '23000')
# - Duplicate entry 'test@gmail.com' for key 'email_address_table.email_address.unique'
# - 1062 (23000): Duplicate entry 'tal@circlez.ai' for key 'person_table.person_table.main_email_person.unique'
# - TODO (index can have any name): 1062 (23000): Duplicate entry '1-2' for key 'test_location_profile_table.idx_location_id_profile_id'
# """
# pattern = r'Duplicate entry \'(.+?)\' for key \'(.+?)\''
# match = re.search(pattern, str(error))
# if not match:  # a different error
#     raise error
# duplicate_value = match.group(1)
# key = match.group(2)
# if key.endswith("PRIMARY"):
#     return int(duplicate_value), ""
# elif key.endswith("_UNIQUE"):
#     # all but last
#     duplicate_column_name = "_".join(key.split(".")[-1].split('_')[:-1])
# elif key.count(".") > 1:
#     duplicate_column_name = key.split(".")[-2]
# else:
#     raise Exception(f"GenericCRUD._get_existing_duplicate_id: please report the following error,"
#                     f" so we can add support to this case: insert error: {error}")
# return duplicate_value, duplicate_column_name

def generate_where_clause_for_ignore_duplicate(data_dict: dict,
                                               constraint_columns: list[list[str]]) -> Tuple[str, tuple]:
    where_clauses = []
    values = tuple()

    for constraint_column in constraint_columns:
        # Handle special case for 'point' or 'coordinate'
        formatted_columns = []
        temp_values = []

        for column in constraint_column:
            if column in data_dict:
                if column == 'point' or column == 'coordinate':
                    formatted_columns.append('ST_X(coordinate) = %s')
                    temp_values.append(data_dict[column])
                    formatted_columns.append('ST_Y(coordinate) = %s')
                    temp_values.append(data_dict[column])
                else:
                    formatted_columns.append(f"{column} = %s")
                    temp_values.append(data_dict[column])

        if (len(temp_values) == len(constraint_column) *
                (2 if 'point' in constraint_column or 'coordinate' in constraint_column else 1)):
            where_clauses.append(f"({' AND '.join(formatted_columns)})")
            values += tuple(temp_values)

    where = ' OR '.join(where_clauses)
    return where, values


# process field with dot in it
def fix_select_clause_value(select_clause_value: str) -> str:
    """If some fields in select_clause_value have "." in them, we have to surround them with ``."""
    # Split the select clause value by commas to handle multiple fields
    if select_clause_value is None:
        return select_clause_value
    fields = select_clause_value.split(',')

    # Process each field
    fixed_fields = []
    for field in fields:
        field = field.strip()
        if '.' in field and field[0] != '`' and field[-1] != '`':
            field = f'`{field}`'
        fixed_fields.append(field)

    # Join the fixed fields back into a single string
    fixed_select_clause_value = ', '.join(fixed_fields)
    return fixed_select_clause_value


def fix_where_clause_value(where_clause_value: str) -> str:
    """If some fields in where_clause_value have "." in them, we have to surround them with ``."""
    # Split the where clause value by AND or OR to handle multiple conditions
    if where_clause_value is None:
        return where_clause_value
    conditions = where_clause_value.split()

    # Process each condition
    fixed_conditions = []
    for condition in conditions:
        condition_column = condition.split('=')[0].strip()  # Get the column name part of the condition
        if '.' in condition_column and condition_column[0] != '`' and condition_column[-1] != '`':
            condition = condition.replace(condition_column, f'`{condition_column}`')  # Surround the column name with `
        fixed_conditions.append(condition)

    # Join the fixed conditions back into a single string
    fixed_where_clause_value = ' '.join(fixed_conditions)
    return fixed_where_clause_value

# TODO: run this with cache decorator if the above get_table_columns doesn't find anything.
# def get_table_columns(schema_name: str = None, table_name: str = None) -> tuple:
#     select_query = "SELECT column_name " \
#                    "FROM information_schema.columns " \
#                    "WHERE TABLE_SCHEMA = %s " \
#                    "AND TABLE_NAME = %s;"
#     params = (schema_name, table_name)
#
#     self.connection.commit()
#     self.cursor.execute(select_query, params)
#     result = self.cursor.fetchall()
#
#     columns_tuple = tuple(x[0] for x in result)
#     self.logger.debug(object=locals())
#     return columns_tuple


#module_wrapper(logger)





from .generic_crud_mysql import GenericCrudMysql


def copy_table_definition(source_schema_name: str, source_table_name: str,
                          target_schema_name: str, target_table_name: str,
                          where_clause: str = None):
    entity_name = source_table_name

    generic_crud = GenericCrudMysql(default_entity_name=entity_name,
                                    default_schema_name=source_schema_name,
                                    default_table_name=source_table_name)
    print(f"copy_table.py copy_table_definition() before CREATE SCHEMA IF NOT EXISTS {target_schema_name}")
    generic_crud.cursor.execute(f"CREATE SCHEMA IF NOT EXISTS {target_schema_name};")
    generic_crud.cursor.execute(f"CREATE TABLE {target_schema_name}.{target_table_name} LIKE {source_schema_name}.{source_table_name};")


def copy_table_data(source_schema_name: str, source_table_name: str,
                    target_schema_name: str, target_table_name: str,
                    where_clause: str = None):
    # TODO we should use get_entity_name()
    entity_name = source_table_name

    generic_crud = GenericCrudMysql(default_entity_name=entity_name,
                                    default_schema_name=source_schema_name,
                                    default_table_name=source_table_name)
    generic_crud.cursor.execute(f"INSERT INTO {target_schema_name}.{target_table_name} SELECT * FROM {source_schema_name}.{source_table_name};")


def copy_table(source_schema_name: str, source_table_name: str,
               target_schema_name: str, target_table_name: str,
               where_clause: str = None):
    copy_table_definition(source_schema_name, source_table_name,
                          target_schema_name, target_table_name, where_clause)
    copy_table_data(source_schema_name, source_table_name, target_schema_name,
                    target_table_name, where_clause)



from database_infrastructure_local.point import Point

# ! this fiel is for backward compatibility
# ! a lot of backeages import point from database_mysql_local,
# ! update them to import from database_infrastructure_local instead then this file can be removed/ deleted
__all__ = ["Point"]



from database_infrastructure_local.to_sql_interface import ToSQLInterface, Now, CurrentDate, TimeStampDiff, Count, TimeUnit, Function, Concat

# ! this file is for backward compatibility, the to_sql_interface.py file has been moved to the database_infrastructure_local package
# ! there is a lot of packages that import from database_mysql_local.to_sql_interface.
# ! Update those imports to database_infrastructure_local.to_sql_interface
# ! this file will be removed in the future

__all__ = [
    "ToSQLInterface",
    "Now",
    "CurrentDate",
    "TimeStampDiff",
    "Count",
    "TimeUnit",
    "Function",
    "Concat"
]



import datetime
import re
from functools import lru_cache
from typing import Optional
from typing import Tuple
from language_remote.lang_code import LangCode
from logger_local.MetaLogger import MetaLogger
from user_context_remote.user_context import UserContext

# PyCham on MacOS need . before
from database_infrastructure_local.constants import DEFAULT_SQL_SELECT_LIMIT

from .version import get_package_version
from .constants_src import CRUD_ML_CODE_LOGGER_OBJECT, IsTestData
from .generic_crud import GenericCRUD
from .table_definition import table_definition
from .utils import generate_id_column_name, generate_table_name, generate_view_name, is_column_in_table

IS_MAIN_COLUMN_NAME = "is_main"

# Use GenericCRUDML only for tables which needs multi-language support


# TODO If I understand, there are two things:
#   - is_main in data_ml_dict
#   - table_definition_table.is_main_column
#   I'm not sure we need additional parameter.
class GenericCRUDML(GenericCRUD, metaclass=MetaLogger, object=CRUD_ML_CODE_LOGGER_OBJECT):
    """A class that provides generic CRUD functionality for tables with multi-language support."""

    # TODO: allow overiding all default values in all methods
    # TODO Shall we add same_entity method as parameter?
    def __init__(self, default_entity_name: str = None,
                 default_schema_name: str = None,
                 default_table_name: str = None,
                 default_view_table_name: str = None,
                 default_ml_table_name: str = None,
                 default_ml_view_table_name: str = None,
                 default_view_with_deleted_and_test_data: str = None,
                 default_column_name: str = None,  # default_select_clause_value: str = "*"
                 is_main_column_name: str = IS_MAIN_COLUMN_NAME, is_test_data: bool = False,
                 is_translate_object_name_per_environment: bool = True, # TODO Shall we add it here if it exists in GenericCRUD?
                 ) -> None:
        """Initializes the GenericCRUDML class. If connection is not provided,
        a new connection will be created."""
        self.default_schema_name = default_schema_name or default_entity_name
        self.default_table_name = default_table_name or generate_table_name(entity_name=default_entity_name,
                                                                            schema_name=default_schema_name,
                                                                            ml_table_name=default_ml_table_name)
        self.default_ml_table_name = default_ml_table_name or self.generate_ml_table_name(default_table_name)
        self.default_ml_view_table_name = default_ml_view_table_name or generate_view_name(self.default_ml_table_name)
        self.default_view_with_deleted_and_test_data = default_view_with_deleted_and_test_data
        self.logger.info("GenericCRUDML.__init__" + get_package_version(), object={
            "default_ml_table_name": self.default_ml_table_name,
            "default_ml_view_table_name": self.default_ml_view_table_name})
        super().__init__(default_entity_name=default_entity_name,
                         default_schema_name=default_schema_name,
                         default_table_name=self.default_table_name,
                         default_view_table_name=default_view_table_name,
                         default_view_with_deleted_and_test_data=default_view_with_deleted_and_test_data,
                         default_column_name=default_column_name,
                         is_test_data=IsTestData.TEMP_DATA.value)
        self.user_context = UserContext()
        # TODO: we may have to change it to a list when we will have multiple is_main columns
        self.is_main_column_name = is_main_column_name

    @lru_cache
    def sql_in_list_by_entity_list_id(self, *, schema_name: str = None, entity_name: str = None,
                                      entity_list_id: int) -> str:
        """Example: select group_id from group.group_list_member_table WHERE group_list_id=1"""
        schema_name = schema_name or self.default_schema_name
        entity_name = entity_name or schema_name
        ids = super().select_multi_dict_by_column_and_value(
            schema_name=schema_name, view_table_name=f"{entity_name}_list_member_view",
            select_clause_value=f"{entity_name}_id",
            column_name=f"{entity_name}_list_id", column_value=entity_list_id)
        if ids:
            result = f" IN ({','.join([str(_id[f'{entity_name}_id']) for _id in ids])})"
        else:
            result = " = TRUE"

        return result

    # TODO: should we rename all those methode similar to CRUD? (select, etc.)
    def add_value(self, *, schema_name: str = None, data_dict: dict = None, data_ml_dict: dict,
                  column_name: str = None,
                  table_id: int = None, lang_code: LangCode = None,
                  is_main: bool | None = None, table_name: str = None,
                  ml_table_name: str = None) -> Optional[tuple]:
        schema_name = schema_name or self.default_schema_name

        lang_code_str = self._get_lang_code_str(lang_code=lang_code, data_ml_dict=data_ml_dict)
        table_name = table_name or self.default_table_name
        # TODO Shall we also add "or self.default_column_name"?
        column_name = column_name or generate_id_column_name(table_name)
        ml_view_name = generate_view_name(ml_table_name) if ml_table_name else self.default_ml_view_table_name
        ml_table_name = ml_table_name or self.default_ml_table_name

        old_cursor = self.cursor
        try:
            self.cursor = self.connection.cursor()
            # if this is the first insert of this data, is_main should be True
            if table_id is None:
                is_main = True
            elif is_main:
                self._update_old_main_value_to_false(
                    schema_name=schema_name, table_name=table_name, table_id=table_id,
                    ml_table_name=ml_table_name)
            # TODO Why we have _column suffix?
            if table_definition.get(ml_table_name, {}).get("is_main_column"):
                data_ml_dict[self.is_main_column_name] = is_main

            # id is the id value of the row in the table_name table
            table_id = table_id or super().insert(
                schema_name=schema_name, table_name=table_name, data_dict=data_dict,
                ignore_duplicate=True, commit_changes=False)
            if table_id is None:  # TODO: delete?
                self.logger.error("Error inserting data_dict", object={"data_dict": data_dict})
                self.connection.rollback()
                return

            if not data_ml_dict:
                self.connection.commit()
                ml_table_id = None
                return table_id, ml_table_id

            data_ml_dict[column_name] = table_id
            data_ml_dict["lang_code"] = lang_code_str

            # ml_id is the id value of the row in the ml_table_name table
            # we do not allow inserting a new value if exist and there's not is_main column
            if not is_main and not table_definition.get(ml_table_name, {}).get("is_main_column"):
                already_exists = super().select_one_value_by_where(
                    schema_name=schema_name, view_table_name=ml_view_name,
                    where=f"{column_name}=%s AND lang_code=%s", params=(table_id, lang_code_str),
                    select_clause_value=column_name)
                if already_exists:
                    raise Exception(f"Value already exists in {ml_table_name} table. Use delete / update.")
            ml_table_id = super().insert(schema_name=schema_name, data_dict=data_ml_dict, table_name=ml_table_name,
                                         ignore_duplicate=True, commit_changes=False)

            self.connection.commit()
        except Exception as e:
            # TODO Shall we add retry mechanism?
            self.connection.rollback()
            raise e
        finally:
            self.cursor.close()
            self.cursor = old_cursor
        return table_id, ml_table_id

    # TODO add parameter if_same_even_deleted_return_existing_id=true - When inserting a deleted entity (all entities except person and tables of codes i.e. country, state, county, city, neighbourhood, street ...) and table_definition.insert_is_undelete=false, we should null all the unique fields of the deleted entity  # noqa
    def add_value_if_not_exist(
            self, *, data_dict: dict = None, data_ml_dict: dict,
            table_id: int = None, lang_code: LangCode = None,
            is_main: bool = None, table_name: str = None, ml_table_name: str = None, schema_name: str = None,
            order_by: str = None) -> Tuple[int, int]:

        schema_name = schema_name or self.default_schema_name
        table_name = table_name or self.default_table_name
        ml_table_name = ml_table_name or self.default_ml_table_name
        column_name = generate_id_column_name(table_name)
        ml_column_name = generate_id_column_name(ml_table_name)

        # check if the row exists in the ml_table
        if data_dict:
            name = data_dict.get("name") or data_dict.get("title")
        else:
            name = data_ml_dict.get("name") or data_ml_dict.get("title")
        # name = data_dict.get("name") or data_ml_dict.get("title")
        title = data_ml_dict.get("title") or data_dict.get("name")
        if not name:
            raise ValueError("name or title is required for add_value_if_not_exist")
        table_id = table_id or self.get_id_by_name(
            schema_name=schema_name, table_name=table_name, ml_table_name=ml_table_name, name=name,
            lang_code=lang_code, column_name=column_name, order_by=order_by)
        ml_table_id = self.get_ml_id_by_title(
            schema_name=schema_name, ml_table_name=ml_table_name, name=title, lang_code=lang_code,
            column_name=ml_column_name, order_by=order_by)
        if not table_id:
            table_id, ml_table_id = self.add_value(
                schema_name=schema_name, data_ml_dict=data_ml_dict, table_id=table_id, lang_code=lang_code,
                is_main=is_main, data_dict=data_dict, table_name=table_name, ml_table_name=ml_table_name,
                column_name=column_name)
        return table_id, ml_table_id

    def update_value_by_id(
            self, *, data_dict: dict = None, data_ml_dict: dict,
            ml_table_id: int, table_id: int, lang_code: LangCode = None,
            is_main: bool = None, table_name: str = None, ml_table_name: str = None, schema_name: str = None,
            limit: int = DEFAULT_SQL_SELECT_LIMIT, order_by: str = None) -> Tuple[int, int]:

        schema_name = schema_name or self.default_schema_name
        lang_code_str = self._get_lang_code_str(lang_code=lang_code, data_ml_dict=data_ml_dict)
        table_name = table_name or self.default_table_name
        ml_table_name = ml_table_name or self.default_ml_table_name
        column_name = generate_id_column_name(table_name)

        if table_id is None:
            raise ValueError("table_id is required for update_value")
        if ml_table_id is None:
            raise ValueError("ml_table_id is required for update_value")
        old_cursor = self.cursor
        try:
            self.cursor = self.connection.cursor()
            if is_main:
                self._update_old_main_value_to_false(schema_name=schema_name, table_name=table_name,
                                                     table_id=table_id, ml_table_name=ml_table_name)
            if table_definition.get(ml_table_name, {}).get("is_main_column"):
                data_ml_dict[self.is_main_column_name] = is_main
            if data_dict:
                super().update_by_column_and_value(schema_name=schema_name, table_name=table_name,
                                                   column_name=column_name,
                                                   column_value=table_id, data_dict=data_dict, limit=limit,
                                                   order_by=order_by,
                                                   commit_changes=False)

            data_ml_dict[column_name] = table_id
            data_ml_dict["lang_code"] = lang_code_str
            column_name = generate_id_column_name(ml_table_name)

            super().update_by_column_and_value(schema_name=schema_name, table_name=ml_table_name,
                                               data_dict=data_ml_dict,
                                               column_name=column_name, column_value=ml_table_id, limit=limit,
                                               order_by=order_by,
                                               commit_changes=False)

            self.connection.commit()
        except Exception as e:
            self.connection.rollback()
            raise e
        finally:
            self.cursor.close()
            self.cursor = old_cursor
        return table_id, ml_table_id

    def upsert_value(
            self, *, data_dict: dict = None, data_ml_dict: dict, data_dict_compare: dict = None,
            table_id: int = None, lang_code: LangCode = None, is_main: bool = None, table_name: str = None,
            ml_table_name: str = None, schema_name: str = None, compare_view_name: str = None,
            limit: int = DEFAULT_SQL_SELECT_LIMIT, order_by: str = None) -> Tuple[int, int]:
        compare_view_name = compare_view_name or self.default_ml_view_table_name

        schema_name = schema_name or self.default_schema_name
        table_name = table_name or self.default_table_name
        ml_table_name = ml_table_name or self.default_ml_table_name
        column_name = generate_id_column_name(table_name)
        ml_column_name = generate_id_column_name(ml_table_name)

        # check if the row exists in the ml_table
        name = data_dict.get("name") or data_ml_dict.get("title")
        title = data_ml_dict.get("title") or data_dict.get("name")
        name_compare = name
        title_to_compare = title
        if not name:
            raise ValueError("name or title is required for upsert_value")
        if data_dict_compare:
            name_compare = data_dict_compare.get("name") or data_dict_compare.get("title")
            title_to_compare = data_dict_compare.get("title") or data_dict_compare.get("name")

        table_id = table_id or self.get_id_by_name(
            schema_name=schema_name, table_name=table_name, ml_table_name=ml_table_name,
            name=name_compare, compare_view_name=compare_view_name,
            column_name=column_name, order_by=order_by)
        ml_table_id = self.get_ml_id_by_title(
            schema_name=schema_name, ml_table_name=ml_table_name, name=title_to_compare, lang_code=lang_code,
            compare_view_name=compare_view_name, column_name=ml_column_name, order_by=order_by)
        if not table_id or not ml_table_id:
            table_id, ml_table_id = self.add_value(
                schema_name=schema_name, data_ml_dict=data_ml_dict, table_id=table_id, lang_code=lang_code,
                is_main=is_main, data_dict=data_dict, table_name=table_name, ml_table_name=ml_table_name,
                column_name=column_name)
        else:
            table_id, ml_table_id = self.update_value_by_id(
                schema_name=schema_name, data_ml_dict=data_ml_dict, ml_table_id=ml_table_id, table_id=table_id,
                lang_code=lang_code, is_main=is_main, data_dict=data_dict, table_name=table_name,
                ml_table_name=ml_table_name, limit=limit, order_by=order_by)
        return table_id, ml_table_id

    def upsert_value_with_abbreviations(
            self, *, data_dict: dict = None, data_ml_dict: dict,

            data_dict_compare: dict = None, table_id: int = None, order_by: str = None,
            table_name: str = None, ml_table_name: str = None, schema_name: str = None, lang_code: LangCode = None,
            compare_view_name: str = None, limit: int = DEFAULT_SQL_SELECT_LIMIT) -> Tuple[int, list[int]]:
        compare_view_name = compare_view_name or self.default_ml_view_table_name
        schema_name = schema_name or self.default_schema_name
        table_name = table_name or self.default_table_name
        ml_table_name = ml_table_name or self.default_ml_table_name
        column_name = generate_id_column_name(table_name)
        ml_column_name = generate_id_column_name(ml_table_name)

        # check if the row exists in the ml_table
        name_and_abbreviations = data_ml_dict.get("title") or data_dict.get("name")
        if not name_and_abbreviations:
            raise ValueError("name or title is required for upsert_value")
        if data_dict_compare:
            name_and_abbreviations_compare = data_dict_compare.get("title") or data_dict_compare.get("name")
        else:
            name_and_abbreviations_compare = name_and_abbreviations

        name = name_and_abbreviations.split(" (")[0]
        name_compare = name_and_abbreviations_compare.split(" (")[0]
        abbreviations = name_and_abbreviations.split(" (")[1][:-1]
        abbreviations_list = abbreviations.split(", ")
        data_ml_dict_with_name = data_ml_dict.copy()
        data_ml_dict_with_name["title"] = name
        data_ml_dict_with_abbreviation_list = []
        for abbreviation in abbreviations_list:
            data_ml_dict_with_abbreviation = data_ml_dict.copy()
            data_ml_dict_with_abbreviation["title"] = abbreviation
            data_ml_dict_with_abbreviation_list.append(data_ml_dict_with_abbreviation)

        table_id = table_id or self.get_id_by_name(
            schema_name=schema_name, table_name=table_name, ml_table_name=ml_table_name,
            name=name_compare, lang_code=lang_code, compare_view_name=compare_view_name,
            column_name=column_name, order_by=order_by)
        ml_table_id = self.get_ml_id_by_title(
            schema_name=schema_name, ml_table_name=ml_table_name, name=name_compare, lang_code=lang_code,
            compare_view_name=compare_view_name, column_name=ml_column_name, order_by=order_by)
        ml_ids_list = []
        if not table_id:
            table_id, ml_table_id = self.add_value(
                schema_name=schema_name, data_ml_dict=data_ml_dict_with_name, table_id=table_id,
                lang_code=lang_code, is_main=True, data_dict=data_dict, table_name=table_name,
                ml_table_name=ml_table_name)
            ml_ids_list.append(ml_table_id)
            for data_ml_dict_with_abbreviation in data_ml_dict_with_abbreviation_list:
                table_id, ml_table_id = self.add_value(
                    schema_name=schema_name, data_ml_dict=data_ml_dict_with_abbreviation, table_id=table_id,
                    lang_code=lang_code, is_main=None, data_dict=data_dict, table_name=table_name,
                    ml_table_name=ml_table_name)
                ml_ids_list.append(ml_table_id)
        else:
            ml_ids_list = list(self.get_ml_ids_by_id(
                schema_name=schema_name, table_id=table_id, ml_table_name=ml_table_name, lang_code=lang_code,
                column_name=column_name, ml_column_name=ml_column_name, compare_view_name=compare_view_name,
                order_by=order_by))
            ml_ids_list.remove(ml_table_id)
            table_id, main_ml_table_id = self.update_value_by_id(
                schema_name=schema_name, data_ml_dict=data_ml_dict_with_name, ml_table_id=ml_table_id,
                table_id=table_id, lang_code=lang_code, is_main=True, data_dict=data_dict,
                table_name=table_name, ml_table_name=ml_table_name, limit=limit, order_by=order_by)
            for index, data_ml_dict_with_abbreviation in enumerate(data_ml_dict_with_abbreviation_list):
                if index < len(ml_ids_list):
                    table_id, ml_table_id = self.update_value_by_id(
                        data_ml_dict=data_ml_dict_with_abbreviation, ml_table_id=ml_ids_list[index],
                        table_id=table_id, lang_code=lang_code, is_main=None, data_dict=data_dict,
                        table_name=table_name, ml_table_name=ml_table_name, schema_name=schema_name,
                        limit=limit, order_by=order_by)
                else:
                    table_id, ml_table_id = self.add_value(
                        schema_name=schema_name, data_ml_dict=data_ml_dict_with_abbreviation, table_id=table_id,
                        lang_code=lang_code, is_main=None, data_dict=data_dict, table_name=table_name,
                        ml_table_name=ml_table_name)
                ml_ids_list.append(ml_table_id)
            ml_ids_list.append(main_ml_table_id)
        return table_id, ml_ids_list

    def get_values_tuple(self, *, table_id: int, schema_name: str = None, ml_view_table_name: str = None,
                         lang_code: LangCode = None, column_name: str = None,
                         select_clause_value: str = "*", order_by: str = None) -> Tuple[int, int]:
        lang_code_str = self._get_lang_code_str(lang_code=lang_code)
        ml_view_table_name = ml_view_table_name or self.default_ml_view_table_name
        column_name = column_name or self.default_column_name
        schema_name = schema_name or self.default_schema_name
        result = super().select_one_tuple_by_where(
            schema_name=schema_name, view_table_name=ml_view_table_name,
            where=f"{column_name}=%s AND lang_code=%s", params=(table_id, lang_code_str),
            select_clause_value=select_clause_value, order_by=order_by)

        return result

    def get_values_dict(self, table_id: int, lang_code: LangCode = None,
                        column_name: str = None, order_by: str = None) -> dict:

        lang_code_str = self._get_lang_code_str(lang_code=lang_code)
        column_name = column_name or generate_id_column_name(self.default_table_name)
        result = super().select_one_dict_by_where(
            view_table_name=self.default_ml_view_table_name, order_by=order_by,
            where=f"{column_name}=%s AND lang_code=%s", params=(table_id, lang_code_str))

        return result

    def get_values_dict_list(self, *, table_id: int, lang_code: LangCode = None,
                             column_name: str = None, order_by: str = None, schema_name: str = None, limit: int = DEFAULT_SQL_SELECT_LIMIT) -> list:
        schema_name = schema_name or self.default_schema_name

        column_name = column_name or self.default_column_name
        lang_code_str = self._get_lang_code_str(lang_code=lang_code)

        result = self.select_multi_dict_by_where(
            schema_name=schema_name, view_table_name=self.default_ml_view_table_name,
            where=f"{column_name}=%s AND lang_code=%s", params=(table_id, lang_code_str),
            order_by=order_by, limit=limit)
        return result

    def get_values_dict_list_by_where(self, *, lang_code: LangCode = None,  # table_id: int,
                                      column_name: str = None, order_by: str = None, schema_name: str = None,
                                      limit: int = DEFAULT_SQL_SELECT_LIMIT, where: str, select_clause_value: str) -> list:
        schema_name = schema_name or self.default_schema_name

        view_table_name = self.get_view_name(schema_name=schema_name)

        column_name = column_name or self.default_column_name
        lang_code_str = self._get_lang_code_str(lang_code=lang_code)

        result = self.select_multi_dict_by_where(
            schema_name=schema_name, view_table_name=view_table_name, select_clause_value=select_clause_value,
            where=f"{where} AND lang_code=%s", params=(lang_code_str,),
            order_by=order_by, limit=limit)
        return result

    def get_main_values_tuple(self, *, table_id: int, column_name: str = None,
                              select_clause_value: str = "*", order_by: str = None, schema_name: str = None) -> Tuple[int, int]:
        schema_name = schema_name or self.default_schema_name

        column_name = column_name or generate_id_column_name(self.default_table_name)
        result = self.select_one_tuple_by_where(
            schema_name=schema_name, view_table_name=self.default_ml_view_table_name,
            where=f"{column_name}=%s AND is_main=True", params=(table_id,),
            select_clause_value=select_clause_value, order_by=order_by)

        return result

    def get_main_values_dict(self, *, table_id: int, column_name: str = None,
                             select_clause_value: str = "*", order_by: str = None, schema_name: str = None) -> dict:
        schema_name = schema_name or self.default_schema_name

        column_name = column_name or self.default_column_name
        result = super().select_one_dict_by_where(
            schema_name=schema_name, view_table_name=self.default_ml_view_table_name,
            where=f"{column_name}=%s AND is_main=True", params=(table_id,),
            select_clause_value=select_clause_value, order_by=order_by)

        return result

    def get_id_by_name(self, *, name: str, schema_name: str = None, table_name: str = None, ml_table_name: str = None,
                       lang_code: LangCode = None,
                       column_name: str = None, order_by: str = None,
                       compare_view_name: str = None) -> Optional[int]:
        schema_name = schema_name or self.default_schema_name

        lang_code_str = self._get_lang_code_str(lang_code=lang_code, name=name)
        table_name = table_name or self.default_table_name
        ml_table_name = ml_table_name or self.default_ml_table_name
        compare_view_name = compare_view_name or generate_view_name(ml_table_name)
        column_name = column_name or generate_id_column_name(table_name)
        if is_column_in_table(table_name=ml_table_name, column_name="title"):
            where = "`title`=%s"
            params = (name,)
            if lang_code:
                where += " AND lang_code=%s"
                params += (lang_code_str,)
        else:
            where = "`name`=%s"
            params = (name,)
            if lang_code:
                where += " AND lang_code=%s"
                params += (lang_code_str,)
        id_by_name_result = super().select_one_value_by_where(
            schema_name=schema_name, view_table_name=compare_view_name,
            where=where, params=params,
            select_clause_value=column_name, order_by=order_by)
        return id_by_name_result

    def get_ml_id_by_title(self, *, name: str, schema_name: str = None, ml_table_name: str = None,
                           lang_code: LangCode = None,
                           column_name: str = None, order_by: str = None,
                           compare_view_name: str = None) -> Optional[int]:

        lang_code_str = self._get_lang_code_str(lang_code=lang_code, name=name)
        ml_table_name = ml_table_name or self.default_ml_table_name
        schema_name = schema_name or self.default_schema_name
        compare_view_name = compare_view_name or generate_view_name(ml_table_name)
        ml_column_name = column_name or generate_id_column_name(ml_table_name)
        if is_column_in_table(table_name=ml_table_name, column_name="title"):
            where = "title=%s"
            params = (name,)
            if lang_code:
                where += " AND lang_code=%s"
                params += (lang_code_str,)
        else:
            where = "`name`=%s"
            params = (name,)
            if lang_code:
                where += " AND lang_code=%s"
                params += (lang_code_str,)
        ml_id_result = super().select_one_value_by_where(
            schema_name=schema_name, view_table_name=compare_view_name,
            select_clause_value=ml_column_name, order_by=order_by,
            where=where, params=(name, lang_code_str))
        return ml_id_result

    def get_ml_ids_by_id(self, table_id: int, ml_table_name: str = None, lang_code: LangCode = None,
                         column_name: str = None, ml_column_name: str = None,
                         order_by: str = None, compare_view_name: str = None, schema_name: str = None) -> tuple:
        schema_name = schema_name or self.default_schema_name

        lang_code_str = None if lang_code is None else lang_code.value
        compare_view_name = compare_view_name or generate_view_name(ml_table_name)
        ml_ids_result = super().select_multi_value_by_where(
            schema_name=schema_name, view_table_name=compare_view_name,
            select_clause_value=ml_column_name, order_by=order_by,
            where=f"{column_name}=%s AND lang_code=%s", params=(table_id, lang_code_str))
        return ml_ids_result

    def get_all_ml_ids_by_id(self, table_id: int, ml_table_name: str = None,
                             column_name: str = None, ml_column_name: str = None,
                             order_by: str = None, compare_view_name: str = None,
                             schema_name: str = None) -> tuple:
        schema_name = schema_name or self.default_schema_name

        compare_view_name = compare_view_name or generate_view_name(ml_table_name)
        ml_ids_result = super().select_multi_value_by_where(
            schema_name=schema_name, view_table_name=compare_view_name,
            select_clause_value=ml_column_name, order_by=order_by,
            where=f"{column_name}=%s", params=(table_id))
        return ml_ids_result

    def delete_by_name(self, *, name: str, lang_code: LangCode = None) -> int:
        lang_code_str = self._get_lang_code_str(lang_code=lang_code, name=name)
        if is_column_in_table(table_name=self.default_ml_view_table_name, column_name="title"):
            where = "title=%s AND lang_code=%s"
        else:
            where = "`name`=%s AND lang_code=%s"
        deleted_id = super().delete_by_where(table_name=self.default_ml_table_name, where=where,
                                             params=(name, lang_code_str))
        return deleted_id

    @staticmethod
    def generate_ml_table_name(table_name: str) -> str:
        if table_name:
            ml_table_name = re.sub(r'(_table)$', '_ml_table', table_name)
            return ml_table_name

    # Change the old row with is_main=True to is_main=None
    def _update_old_main_value_to_false(self, *, schema_name: str, table_id: int, table_name: str,
                                        ml_table_name: str) -> int:
        data_ml_dict = {self.is_main_column_name: False}
        column_name = generate_id_column_name(table_name)
        where = f"{column_name}=%s AND {self.is_main_column_name} = True"
        updated_id = super().update_by_where(schema_name=schema_name, table_name=ml_table_name,
                                             data_dict=data_ml_dict, where=where, params=(table_id,))
        return updated_id

    def _get_lang_code_str(self, *, name: str = None, lang_code: LangCode = None,
                           data_ml_dict: dict = None) -> str:

        if lang_code is None and name is not None:
            lang_code_str = LangCode.detect_lang_code_str(name)
        elif lang_code is None and data_ml_dict is not None:
            text = data_ml_dict.get("title") or data_ml_dict.get("name")
            lang_code_str = LangCode.detect_lang_code_str(text)
        elif lang_code is None:
            lang_code_str = self.user_context.get_effective_profile_preferred_lang_code_string()
        else:
            lang_code_str = lang_code.value
        return lang_code_str

    def merge_ml_entities(self, entity_id1: int, entity_id2: int, main_entity_ml_id: int = None):
        DEFAULT_ORDER_BY = "updated_timestamp DESC"
        schema_name = self.default_schema_name
        table_name = self.default_table_name
        table_prefix = table_name.replace("_ml_table", "")
        generated_ml_table_name = table_prefix + '_ml_table'
        generated_ml_view_name = table_prefix + '_ml_view'
        generated_table_name = table_prefix + '_table'
        gcrml = GenericCRUDML(default_schema_name=schema_name,
                              default_ml_table_name=generated_ml_table_name,
                              is_test_data=IsTestData.TEMP_DATA.value,
                              default_table_name=generated_table_name,
                              default_ml_view_table_name=generated_ml_view_name,
                              is_main_column_name='is_main_title')
        # establish which id is being merged/ended and which one it is being merged into
        end_id = entity_id1
        main_id = entity_id2

        # Data to update
        old_id = end_id
        new_id = main_id
        # get the ml ids of the city id which is being merged

        generated_ml_column_name = table_prefix + '_ml_id'
        generated_column_name = table_prefix + '_id'
        ml_ids = gcrml.get_all_ml_ids_by_id(table_id=old_id, ml_column_name=generated_ml_column_name, order_by=DEFAULT_ORDER_BY,
                                            ml_table_name=generated_ml_table_name,
                                            column_name=generated_column_name, compare_view_name=generated_ml_view_name)

        # get a list of existing langcodes for the id which is being merged into
        langcode_list = self.get_lang_codes_by_table_id(table_id=new_id, ml_view_table_name=generated_ml_view_name,
                                                        schema_name=self.default_schema_name)

        for id in ml_ids:
            # get the langcode of the ml id being merged. if it already exists in the id being merged into, then set this one to false.
            langcode = gcrml.select_one_value_by_column_and_value(select_clause_value='lang_code', column_name=generated_ml_column_name,
                                                                  column_value=id, view_table_name=generated_ml_view_name)
            if langcode in langcode_list:
                gcrml.update_by_column_and_value(
                    table_name=generated_ml_table_name,
                    column_name=generated_ml_column_name,
                    column_value=id,
                    data_dict={'is_main_title': None}, )
            # update the id for the ml_id to the new id
            gcrml.update_by_column_and_value(
                table_name=generated_ml_table_name,
                column_name=generated_ml_column_name,
                column_value=id,
                data_dict={generated_column_name: new_id, }
            )

        gcrml.update_by_column_and_value(
            table_name=generated_table_name,
            column_name=generated_column_name,
            column_value=old_id,
            data_dict={"end_timestamp": datetime.datetime.now()},
        )
        # Set the main id
        if main_entity_ml_id is not None:
            gcrml.update_by_column_and_value(
                table_name=generated_ml_table_name,
                column_name=generated_ml_column_name,
                column_value=main_entity_ml_id,
                data_dict={'is_main_title': True},)

    def get_lang_codes_by_table_id(self, table_id: int, schema_name: str = None, ml_view_table_name: str = None) -> list:
        schema_name = schema_name or self.default_schema_name
        ml_view_table_name = ml_view_table_name or self.default_ml_view_table_name
        table_name = self.default_table_name
        table_prefix = table_name.replace("_ml_table", "")
        generated_column_name = table_prefix + '_id'

        result = self.select_multi_dict_by_where(
            schema_name=schema_name,
            view_table_name=ml_view_table_name,
            where=f"{generated_column_name}=%s",
            params=(table_id,),
            select_clause_value="lang_code"
        )

        return [entry['lang_code'] for entry in result if 'lang_code' in entry]





from .generic_crud import GenericCRUD

# GenericCrudMysql is a replacement of GenericCRUD for MySQL specific implementations


class GenericCrudMysql(GenericCRUD):
    # def __init__(self, *, default_schema_name: str,
    #              default_table_name: str = None,
    #              default_view_table_name: str = None,
    #              default_view_with_deleted_and_test_data: str = None,
    #              default_column_name: str = None,
    #              default_select_clause_value: str = "*",
    #              default_where: str = None, is_test_data: bool = False) -> None:
    #     # TODO I'm not sure this is the right solution for backward compatability, I think we should rename the existing class to the new name and create new class with the original name.
    #     super()
    pass



from logger_local.MetaLogger import MetaLogger

from .constants_src import MAPPING_CODE_LOGGER_OBJECT
from .generic_crud import GenericCRUD


class GenericMapping(GenericCRUD, metaclass=MetaLogger, object=MAPPING_CODE_LOGGER_OBJECT):
    def __init__(self, default_schema_name: str = None,
                 default_table_name: str = None,
                 default_view_table_name: str = None,
                 default_view_with_deleted_and_test_data: str = None,
                 default_column_name: str = None,
                 default_entity_name1: str = None,
                 default_entity_name2: str = None,
                 is_test_data: bool = False):
        super().__init__(default_schema_name=default_schema_name,
                         default_table_name=default_table_name, default_view_table_name=default_view_table_name,
                         default_view_with_deleted_and_test_data=default_view_with_deleted_and_test_data,
                         default_column_name=default_column_name, is_test_data=is_test_data)
        self.default_entity_name1 = default_entity_name1
        self.default_entity_name2 = default_entity_name2

    # We added the schema_name parameter to avoid using USE and void creating a new instance/object
    # of the database for mapping.
    def insert_mapping(self, *, entity_id1: int, entity_id2: int,
                       entity_name1: str = None, entity_name2: str = None,
                       schema_name: str = None, data_dict: dict = None,
                       ignore_duplicate: bool = False) -> int:
        """Inserts a new link between two entities and returns the id of the
          new row or -1 if an error occurred.
        :param entity_name1: The name of the first entity's table.
        :param entity_name2: The name of the second entity's table.
        :param entity_id1: The id of the first entity.
        :param entity_id2: The id of the second entity.
        :param schema_name: The name of the schema.
        :param data_dict: The data to insert.
        :param ignore_duplicate: If True, ignore duplicate rows.
        :return: The id of the new row or -1 if an error occurred.
        """
        entity_name1 = entity_name1 or self.default_entity_name1
        entity_name2 = entity_name2 or self.default_entity_name2
        schema_name = schema_name or self.default_schema_name
        table_name = f"{entity_name1}_{entity_name2}_table"
        data_dict = data_dict or {}
        data_dict[f"{entity_name1}_id"] = entity_id1
        data_dict[f"{entity_name2}_id"] = entity_id2

        mapping_id = super().insert(schema_name=schema_name, data_dict=data_dict, table_name=table_name,
                                    ignore_duplicate=ignore_duplicate)
        return mapping_id

    # The migration from generic_crud.insert() to generic_mapping.insert_mapping_if_not_exists() should be easy
    # We need to change table_name parameter to entity_ids, entity_names, sometime remove the data_dict?
    def insert_mapping_if_not_exists(self, *, entity_id1: int, entity_id2: int,
                                     entity_name1: str = None, entity_name2: str = None,
                                     schema_name: str = None, data_dict: dict = None,
                                     view_table_name: str = None, data_dict_compare: dict = None) -> int:
        """Inserts a new link between two entities if it does not exist and returns the id of the
            new row or -1 if an error occurred.
        :param entity_name1: The name of the first entity's table.
        :param entity_name2: The name of the second entity's table.
        :param entity_id1: The id of the first entity.
        :param entity_id2: The id of the second entity.
        :param schema_name: The name of the schema.
        :param data_dict: The data to insert.
        :return: The id of the new row or -1 if an error occurred.
        """
        data_dict = data_dict or {}
        data_dict_compare = data_dict_compare or {}
        entity_name1 = entity_name1 or self.default_entity_name1
        entity_name2 = entity_name2 or self.default_entity_name2
        schema_name = schema_name or self.default_schema_name
        table_name = f"{entity_name1}_{entity_name2}_table"
        view_table_name = view_table_name or f"{entity_name1}_{entity_name2}_view"
        data_dict_compare[f"{entity_name1}_id"] = data_dict[f"{entity_name1}_id"] = entity_id1
        data_dict_compare[f"{entity_name2}_id"] = data_dict[f"{entity_name2}_id"] = entity_id2
        mapping_id = super().insert_if_not_exists(schema_name=schema_name, data_dict=data_dict, table_name=table_name,
                                                  data_dict_compare=data_dict_compare, view_table_name=view_table_name)
        return mapping_id

    def upsert_mapping(self, *, entity_id1: int, entity_id2: int,
                       entity_name1: str = None, entity_name2: str = None,
                       schema_name: str = None, data_dict: dict = None,
                       view_table_name: str = None, data_dict_compare: dict = None) -> int:
        """Inserts a new link between two entities and returns the id of the
            new row or -1 if an error occurred.
        :param entity_name1: The name of the first entity's table.
        :param entity_name2: The name of the second entity's table.
        :param entity_id1: The id of the first entity.
        :param entity_id2: The id of the second entity.
        :param schema_name: The name of the schema.
        :param data_dict: The data to insert.
        :return: The id of the new row or -1 if an error occurred.
        """
        data_dict = data_dict or {}
        data_dict_compare = data_dict_compare or {}
        entity_name1 = entity_name1 or self.default_entity_name1
        entity_name2 = entity_name2 or self.default_entity_name2
        schema_name = schema_name or self.default_schema_name
        table_name = f"{entity_name1}_{entity_name2}_table"
        data_dict_compare[f"{entity_name1}_id"] = data_dict[f"{entity_name1}_id"] = entity_id1
        data_dict_compare[f"{entity_name2}_id"] = data_dict[f"{entity_name2}_id"] = entity_id2
        mapping_id = super().upsert(schema_name=schema_name, data_dict=data_dict, table_name=table_name,
                                    view_table_name=view_table_name, data_dict_compare=data_dict_compare)
        return mapping_id

    def insert_mapping_with_ml_ids(self, *, entity_id1: int, entity_id2: int,
                                   entity_name1: str = None, entity_name2: str = None,
                                   entity_ml_ids_list1: list[int] = None, entity_ml_ids_list2: list[int] = None,
                                   schema_name: str = None, data_dict: dict = None,
                                   ignore_duplicate: bool = False) -> list[int]:
        """
        Inserts a new link between two entities and returns the list of mapping ids inserted.
        :param entity_name1: The name of the first entity's table.
        :param entity_name2: The name of the second entity's table.
        :param entity_id1: The id of the first entity.
        :param entity_id2: The id of the second entity.
        :param entity_ml_ids_list1: The list of the first entity's multi language ids.
        :param entity_ml_ids_list2: The list of the second entity's multi language ids.
        :param schema_name: The name of the schema.
        :param data_dict: The data to insert.
        :param ignore_duplicate: If True, ignore duplicate rows.
        :return: List of mapping ids.
        """
        data_dict = data_dict or {}
        mapping_ids_list: list[int] = []
        if entity_ml_ids_list1:
            self.logger.info("insert_mapping_with_ml_ids: entity_ml_ids_list1 is not empty")
            for entity_ml_id1 in entity_ml_ids_list1:
                if entity_ml_ids_list2:
                    for entity_ml_id2 in entity_ml_ids_list2:
                        self.logger.info(
                            "insert_mapping_with_ml_ids: entity_ml_ids_list1 and entity_ml_ids_list2 are not empty")
                        data_dict[f"{entity_name1}_ml_id"] = entity_ml_id1
                        data_dict[f"{entity_name2}_ml_id"] = entity_ml_id2
                        mapping_id = self.insert_mapping(
                            entity_id1=entity_id1, entity_id2=entity_id2,
                            entity_name1=entity_name1, entity_name2=entity_name2,
                            schema_name=schema_name, data_dict=data_dict,
                            ignore_duplicate=ignore_duplicate)
                        self.logger.info(
                            "insert_mapping_with_ml_ids: inserted mapping",
                            object={"mapping_id": mapping_id, "entity_name1": entity_name1, "entity_name2": entity_name2,
                                    "entity_ml_id1": entity_ml_id1, "entity_ml_id2": entity_ml_id2})
                        mapping_ids_list.append(mapping_id)
                else:
                    self.logger.info(
                        "insert_mapping_with_ml_ids: entity_ml_ids_list1 is not empty and entity_ml_ids_list2 is empty")
                    data_dict[f"{entity_name1}_ml_id"] = entity_ml_id1
                    mapping_id = self.insert_mapping(
                        entity_id1=entity_id1, entity_id2=entity_id2,
                        entity_name1=entity_name1, entity_name2=entity_name2,
                        schema_name=schema_name, data_dict=data_dict,
                        ignore_duplicate=ignore_duplicate)
                    self.logger.info(
                        "insert_mapping_with_ml_ids: inserted mapping",
                        object={"mapping_id": mapping_id, "entity_name1": entity_name1, "entity_name2": entity_name2,
                                "entity_ml_id1": entity_ml_id1, "entity_ml_id2": None})
                    mapping_ids_list.append(mapping_id)
        elif entity_ml_ids_list2:
            self.logger.info(
                "insert_mapping_with_ml_ids: entity_ml_ids_list1 is empty and entity_ml_ids_list2 is not empty")
            for entity_ml_id2 in entity_ml_ids_list2:
                data_dict[f"{entity_name2}_ml_id"] = entity_ml_id2
                mapping_id = self.insert_mapping(
                    entity_id1=entity_id1, entity_id2=entity_id2,
                    entity_name1=entity_name1, entity_name2=entity_name2,
                    schema_name=schema_name, data_dict=data_dict,
                    ignore_duplicate=ignore_duplicate)
                self.logger.info(
                    "insert_mapping_with_ml_ids: inserted mapping",
                    object={"mapping_id": mapping_id, "entity_name1": entity_name1, "entity_name2": entity_name2,
                            "entity_ml_id1": None, "entity_ml_id2": entity_ml_id2})
                mapping_ids_list.append(mapping_id)
        else:
            self.logger.info(
                "insert_mapping_with_ml_ids: entity_ml_ids_list1 and entity_ml_ids_list2 are empty")
            mapping_id = self.insert_mapping(
                entity_id1=entity_id1, entity_id2=entity_id2,
                entity_name1=entity_name1, entity_name2=entity_name2,
                schema_name=schema_name, data_dict=data_dict,
                ignore_duplicate=ignore_duplicate)
            self.logger.info(
                "insert_mapping_with_ml_ids: inserted mapping",
                object={"mapping_id": mapping_id, "entity_name1": entity_name1, "entity_name2": entity_name2,
                        "entity_ml_id1": None, "entity_ml_id2": None})
            mapping_ids_list.append(mapping_id)
        return mapping_ids_list

    def insert_mapping_if_not_exists_with_ml_ids(self, *, entity_id1: int, entity_id2: int,
                                                 entity_name1: str = None, entity_name2: str = None,
                                                 entity_ml_ids_list1: list[int] = None, entity_ml_ids_list2: list[int] = None,
                                                 schema_name: str = None, data_dict: dict = None,
                                                 view_table_name: str = None, data_dict_compare: dict = None) -> list[int]:
        """
        Inserts a new link between two entities if it does not exist and returns the list of mapping ids inserted.
        :param entity_name1: The name of the first entity's table.
        :param entity_name2: The name of the second entity's table.
        :param entity_id1: The id of the first entity.
        :param entity_id2: The id of the second entity.
        :param entity_ml_ids_list1: The list of the first entity's multi language ids.
        :param entity_ml_ids_list2: The list of the second entity's multi language ids.
        :param schema_name: The name of the schema.
        :param data_dict: The data to insert.
        :param data_dict_compare: The data to compare to check if the mapping already exists in the db.
        :param view_table_name: The name of the view in which we compare with data_dict_compare.
        :return: List of mapping ids.
        """
        data_dict = data_dict or {}
        data_dict_compare = data_dict_compare or {}
        mapping_ids_list: list[int] = []
        if entity_ml_ids_list1:
            self.logger.info(
                "insert_mapping_if_not_exists_with_ml_ids: entity_ml_ids_list1 is not empty")
            for entity_ml_id1 in entity_ml_ids_list1:
                if entity_ml_ids_list2:
                    for entity_ml_id2 in entity_ml_ids_list2:
                        self.logger.info(
                            "insert_mapping_if_not_exists_with_ml_ids: entity_ml_ids_list1 and entity_ml_ids" +
                            "_list2 are not empty")
                        data_dict[f"{entity_name1}_ml_id"] = entity_ml_id1
                        data_dict[f"{entity_name2}_ml_id"] = entity_ml_id2
                        data_dict_compare[f"{entity_name1}_ml_id"] = entity_ml_id1
                        data_dict_compare[f"{entity_name2}_ml_id"] = entity_ml_id2
                        mapping_id = self.insert_mapping_if_not_exists(
                            entity_id1=entity_id1, entity_id2=entity_id2,
                            entity_name1=entity_name1, entity_name2=entity_name2,
                            schema_name=schema_name, data_dict=data_dict,
                            view_table_name=view_table_name, data_dict_compare=data_dict_compare)
                        self.logger.info(
                            "insert_mapping_if_not_exists_with_ml_ids: inserted mapping",
                            object={"mapping_id": mapping_id, "entity_name1": entity_name1, "entity_name2": entity_name2,
                                    "entity_ml_id1": entity_ml_id1, "entity_ml_id2": entity_ml_id2})
                        mapping_ids_list.append(mapping_id)
                else:
                    self.logger.info(
                        "insert_mapping_if_not_exists_with_ml_ids: entity_ml_ids_list1 is not empty and entity_ml_ids_list2 is empty")
                    data_dict[f"{entity_name1}_ml_id"] = entity_ml_id1
                    data_dict_compare[f"{entity_name1}_ml_id"] = entity_ml_id1
                    mapping_id = self.insert_mapping_if_not_exists(
                        entity_id1=entity_id1, entity_id2=entity_id2,
                        entity_name1=entity_name1, entity_name2=entity_name2,
                        schema_name=schema_name, data_dict=data_dict,
                        view_table_name=view_table_name, data_dict_compare=data_dict_compare)
                    self.logger.info(
                        "insert_mapping_if_not_exists_with_ml_ids: inserted mapping",
                        object={"mapping_id": mapping_id, "entity_name1": entity_name1, "entity_name2": entity_name2,
                                "entity_ml_id1": entity_ml_id1, "entity_ml_id2": None})
                    mapping_ids_list.append(mapping_id)
        elif entity_ml_ids_list2:
            self.logger.info(
                "insert_mapping_if_not_exists_with_ml_ids: entity_ml_ids_list1 is empty and entity_ml_ids_list2 is not empty")
            for entity_ml_id2 in entity_ml_ids_list2:
                data_dict[f"{entity_name2}_ml_id"] = entity_ml_id2
                data_dict_compare[f"{entity_name2}_ml_id"] = entity_ml_id2
                mapping_id = self.insert_mapping_if_not_exists(
                    entity_id1=entity_id1, entity_id2=entity_id2,
                    entity_name1=entity_name1, entity_name2=entity_name2,
                    schema_name=schema_name, data_dict=data_dict,
                    view_table_name=view_table_name, data_dict_compare=data_dict_compare)
                self.logger.info(
                    "insert_mapping_if_not_exists_with_ml_ids: inserted mapping",
                    object={"mapping_id": mapping_id, "entity_name1": entity_name1, "entity_name2": entity_name2,
                            "entity_ml_id1": None, "entity_ml_id2": entity_ml_id2})
                mapping_ids_list.append(mapping_id)
        else:
            self.logger.info(
                "insert_mapping_if_not_exists_with_ml_ids: entity_ml_ids_list1 and entity_ml_ids_list2 are empty")
            mapping_id = self.insert_mapping_if_not_exists(
                entity_id1=entity_id1, entity_id2=entity_id2,
                entity_name1=entity_name1, entity_name2=entity_name2,
                schema_name=schema_name, data_dict=data_dict,
                view_table_name=view_table_name)
            self.logger.info(
                "insert_mapping_if_not_exists_with_ml_ids: inserted mapping",
                object={"mapping_id": mapping_id, "entity_name1": entity_name1, "entity_name2": entity_name2,
                        "entity_ml_id1": None, "entity_ml_id2": None})
            mapping_ids_list.append(mapping_id)
        return mapping_ids_list

    def upsert_mapping_with_ml_ids(self, *, entity_id1: int, entity_id2: int,
                                   entity_name1: str = None, entity_name2: str = None,
                                   entity_ml_ids_list1: list[int] = None, entity_ml_ids_list2: list[int] = None,
                                   schema_name: str = None, data_dict: dict = None,
                                   view_table_name: str = None, data_dict_compare: dict = None) -> list[int]:
        data_dict = data_dict or {}
        data_dict_compare = data_dict_compare or {}
        mapping_ids_list: list[int] = []
        if entity_ml_ids_list1:
            self.logger.info(
                "upsert_mapping_with_ml_ids: entity_ml_ids_list1 is not empty")
            for entity_ml_id1 in entity_ml_ids_list1:
                if entity_ml_ids_list2:
                    for entity_ml_id2 in entity_ml_ids_list2:
                        self.logger.info(
                           "upsert_mapping_with_ml_ids: entity_ml_ids_list1 and entity_ml_ids_list2 are not empty")
                        data_dict[f"{entity_name1}_ml_id"] = entity_ml_id1
                        data_dict[f"{entity_name2}_ml_id"] = entity_ml_id2
                        data_dict_compare[f"{entity_name1}_ml_id"] = entity_ml_id1
                        data_dict_compare[f"{entity_name2}_ml_id"] = entity_ml_id2
                        mapping_id = self.upsert_mapping(
                            entity_id1=entity_id1, entity_id2=entity_id2,
                            entity_name1=entity_name1, entity_name2=entity_name2,
                            schema_name=schema_name, data_dict=data_dict,
                            view_table_name=view_table_name, data_dict_compare=data_dict_compare)
                        self.logger.info(
                            "upsert_mapping_with_ml_ids: inserted mapping",
                            object={"mapping_id": mapping_id, "entity_name1": entity_name1, "entity_name2": entity_name2,
                                    "entity_ml_id1": entity_ml_id1, "entity_ml_id2": entity_ml_id2})
                        mapping_ids_list.append(mapping_id)
                else:
                    self.logger.info(
                        "upsert_mapping_with_ml_ids: entity_ml_ids_list1 is not empty and entity_ml_ids_list2 is empty")
                    data_dict[f"{entity_name1}_ml_id"] = entity_ml_id1
                    data_dict_compare[f"{entity_name1}_ml_id"] = entity_ml_id1
                    mapping_id = self.upsert_mapping(
                        entity_id1=entity_id1, entity_id2=entity_id2,
                        entity_name1=entity_name1, entity_name2=entity_name2,
                        schema_name=schema_name, data_dict=data_dict,
                        view_table_name=view_table_name, data_dict_compare=data_dict_compare)
                    self.logger.info(
                        "upsert_mapping_with_ml_ids: inserted mapping",
                        object={"mapping_id": mapping_id, "entity_name1": entity_name1, "entity_name2": entity_name2,
                                "entity_ml_id1": entity_ml_id1, "entity_ml_id2": None})
                    mapping_ids_list.append(mapping_id)
        elif entity_ml_ids_list2:
            self.logger.info(
                "upsert_mapping_with_ml_ids: entity_ml_ids_list1 is empty and entity_ml_ids_list2 is not empty")
            for entity_ml_id2 in entity_ml_ids_list2:
                data_dict[f"{entity_name2}_ml_id"] = entity_ml_id2
                data_dict_compare[f"{entity_name2}_ml_id"] = entity_ml_id2
                mapping_id = self.upsert_mapping(
                    entity_id1=entity_id1, entity_id2=entity_id2,
                    entity_name1=entity_name1, entity_name2=entity_name2,
                    schema_name=schema_name, data_dict=data_dict,
                    view_table_name=view_table_name, data_dict_compare=data_dict_compare)
                self.logger.info(
                    "upsert_mapping_with_ml_ids: inserted mapping",
                    object={"mapping_id": mapping_id, "entity_name1": entity_name1, "entity_name2": entity_name2,
                            "entity_ml_id1": None, "entity_ml_id2": entity_ml_id2})
                mapping_ids_list.append(mapping_id)
        else:
            self.logger.info(
                "upsert_mapping_with_ml_ids: entity_ml_ids_list1 and entity_ml_ids_list2 are empty")
            mapping_id = self.upsert_mapping(
                entity_id1=entity_id1, entity_id2=entity_id2,
                entity_name1=entity_name1, entity_name2=entity_name2,
                schema_name=schema_name, data_dict=data_dict,
                view_table_name=view_table_name, data_dict_compare=data_dict_compare)
            self.logger.info(
                "upsert_mapping_with_ml_ids: inserted mapping",
                object={"mapping_id": mapping_id, "entity_name1": entity_name1, "entity_name2": entity_name2,
                        "entity_ml_id1": None, "entity_ml_id2": None})
            mapping_ids_list.append(mapping_id)
        return mapping_ids_list

        # TODO: do we need delete_mapping_by_id?

    def delete_mapping_by_two_ids(self, *, entity_id1: int, entity_id2: int,
                                  entity_name1: str = None, entity_name2: str = None,
                                  schema_name: str = None) -> None:
        """ Deletes a link between two entities.
        :param entity_name1: The name of the first entity's table.
        :param entity_name2: The name of the second entity's table.
        :param entity_id1: The id of the first entity.
        :param entity_id2: The id of the second entity.
        :param schema_name: The name of the schema.
        :return: None
        """

        entity_name1 = entity_name1 or self.default_entity_name1
        entity_name2 = entity_name2 or self.default_entity_name2
        schema_name = schema_name or self.default_schema_name
        table_name = f"{entity_name1}_{entity_name2}_table"

        where = f"{entity_name1}_id=%s AND {entity_name2}_id=%s"
        params = (entity_id1, entity_id2)
        super().delete_by_where(schema_name=schema_name, table_name=table_name, where=where, params=params)

    # TODO: remove _by_id, as there's no by_where. Plus, it's by ids. make it backward compatible
    def select_multi_mapping_tuple_by_id(self, *, entity_id1: int, entity_id2: int,
                                         entity_name1: str = None, entity_name2: str = None,
                                         schema_name: str = None, select_clause_value: str = "*") -> list:
        """Selects a row from the database by id.
        :param entity_name1: The name of the first entity's table.
        :param entity_name2: The name of the second entity's table.
        :param entity_id1: The id of the first entity.
        :param entity_id2: The id of the second entity.
        :param schema_name: The name of the schema.
        :param select_clause_value: The columns to select.
        :return: A list of dictionaries representing the rows.
        """
        assert entity_id1 is not None
        assert entity_id2 is not None

        entity_name1 = entity_name1 or self.default_entity_name1
        entity_name2 = entity_name2 or self.default_entity_name2
        schema_name = schema_name or self.default_schema_name
        table_name = f"{entity_name1}_{entity_name2}_view"
        where = f"{entity_name1}_id=%s AND {entity_name2}_id=%s"
        params = (entity_id1, entity_id2)
        result = super().select_multi_tuple_by_where(schema_name=schema_name, view_table_name=table_name,
                                                     select_clause_value=select_clause_value,
                                                     where=where, params=params)
        return result

    def select_multi_mapping_dict_by_id(self, *, entity_id1: int, entity_id2: int,
                                        entity_name1: str = None, entity_name2: str = None,
                                        schema_name: str = None, select_clause_value: str = "*") -> list:
        """Selects a row from the database by id.
        :param entity_name1: The name of the first entity's table.
        :param entity_name2: The name of the second entity's table.
        :param entity_id1: The id of the first entity.
        :param entity_id2: The id of the second entity.
        :param schema_name: The name of the schema.
        :param select_clause_value: The columns to select.
        :return: A list of dictionaries representing the rows.
        """

        result = self.select_multi_mapping_tuple_by_id(entity_name1=entity_name1, entity_name2=entity_name2,
                                                       entity_id1=entity_id1, entity_id2=entity_id2,
                                                       schema_name=schema_name,
                                                       select_clause_value=select_clause_value)

        result = super().convert_multi_to_dict(result, select_clause_value)
        return result

    # TODO: add select_one_dict/tuple/value, and select_multi_value





# generate_table_columns.py

import json

from src.connector import Connector

# TODO improve Sql2Code. We prefer only Sql2Code to generate those structures.

connection = Connector.connect(schema_name="INFORMATION_SCHEMA")
try:
    tables_and_columns = {}
    cursor = connection.cursor()
    query = """SELECT TABLE_NAME, COLUMN_NAME
                FROM INFORMATION_SCHEMA.COLUMNS
                WHERE TABLE_SCHEMA <> 'information_schema'"""

    # Execute the query
    cursor.execute(query)
    result = cursor.fetchall()

    # Process the results and group by table name
    for row in result:
        table_name, column_name = row
        if table_name not in tables_and_columns:
            tables_and_columns[table_name] = []
        tables_and_columns[table_name].append(column_name)
        print(f"Table: {table_name}, Column: {column_name}")

    print(f"\n {len(tables_and_columns)} tables found (the minimum should be at least 1332).\n")
    print(f"{sum(len(columns) for columns in tables_and_columns.values())} columns found (the minimun should be at least 16122).\n")

    table_columns_json = json.dumps(tables_and_columns, indent=4).replace("[", "(").replace("]", ")")
    with open("src/table_columns.py", "w") as f:
        # TODO Add date time to the 1st f.write()
        f.write("# Generated by generate_table_columns.py from MySQL system tables. unlike table_definition which generated from our applicative table in the database (should be generated by Sql2Code in the future\n\n")
        f.write(f"table_columns = {table_columns_json}\n")
    print("Table columns have been generated and saved to table_columns.py")
finally:
    connection.close()



from python_sdk_remote.our_object import OurObject


class HelperMysqlObject(OurObject):
    # use this as test class, naming it TestMysqlObject results in a Warning:
    # cannot collect test class 'TestMysqlObject' because it has a __init__ constructor (from: database_mysql_local/src/test_mysql_object.py)
    def __init__(self, **kwargs):
        test_mysql_fields = {
            "test_mysql_id",
            "number",
            "identifier",
            "name",
            "group",
            "unique_int_column",
            "point",
            "polygon",
            "visibility_id",
            "is_test_data",
            "start_timestamp",
            "end_timestamp",
            "created_timestamp",
            "created_user_id",
            "created_real_user_id",
            "created_effective_user_id",
            "created_effective_profile_id",
            "updated_timestamp",
            "updated_user_id",
            "updated_real_user_id",
            "updated_effective_user_id",
            "updated_effective_profile_id",
        }

        for field, value in kwargs.items():
            if field in test_mysql_fields:
                setattr(self, field, value)

    def get_name(self):
        return self.name

    def __str__(self):
        return "test_mysql: " + str(self.__dict__)

    def __repr__(self):
        return "test_mysql: " + str(self.__dict__)

    def to_dict(self):
        return self.__dict__



import setuptools


PACKAGE_NAME = "database-mysql-local"
package_dir = PACKAGE_NAME.replace("-", "_")

setuptools.setup(
    name=PACKAGE_NAME,  # https://pypi.org/project/database-mysql-local
    version='0.1.54b2868',
    author="Circles",
    author_email="info@circlez.ai",
    url=f"https://github.com/circles-zone/{PACKAGE_NAME}-python-package",
    packages=[package_dir],
    package_dir={package_dir: f'{package_dir}/src'},
    package_data={package_dir: ['*.py']},
    long_description="Database MySQL Local ",
    long_description_content_type='text/markdown',
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        # https://pypi.org/project/mysql-connector-python/
        "mysql-connector-python>=9.0.0",
        "url-remote>=0.0.119",  # https://pypi.org/project/url-remote/
        # TODO Do not use Beta version
        "logger-local>=0.0.186b2614",  # https://pypi.org/project/logger-local/
        # database-infrastructure-local 0.0.35 generic_crud_abstract.py GenericCrudAbstract
        # https://pypi.org/project/database-infrastructure-local/
        "database-infrastructure-local>=0.1.4",
        "language-remote>=0.0.24",  # https://pypi.org/project/language-remote/
        "sql-to-code-local>=0.0.18",  # https://pypi.org/project/sql-to-code-local/
        "python-sdk-remote>=0.0.153",
        # Commented because of a problem with serverless, TODO We should add a test to spot such case
        # "sshtunnel>=0.4.0", # https://pypi.org/project/sshtunnel/
    ]
)



from typing import Any

from logger_local.MetaLogger import MetaLogger
from mysql.connector.cursor import MySQLCursor
from mysql import connector
from .constants_src import LOGGER_CONNECTOR_CODE_OBJECT
# from python_sdk_remote.utilities import our_get_env

# MYSQL_VERSION = our_get_env("MYSQL_VERSION", raise_if_not_found=False)
# TODO Is it the MySQL engine version or the connector version?
MYSQL_VERSION = connector.__version__


# TODO Move this function to a method in DatabaseMysql class. Add this method to the database abstract class
def version_tuple(version):
    return tuple(map(int, version.split(".")))  # Converts "9.2.0" → (9, 2, 0)


class Cursor(metaclass=MetaLogger, object=LOGGER_CONNECTOR_CODE_OBJECT):
    def __init__(self, cursor: MySQLCursor) -> None:
        self.cursor = cursor
        self.__is_closed = False

    # TODO: If environment <> prod1 and dvlp1 break down using 3rd party package and analyze the formatted_sql
    #  and call private method _validate_select_table_name(table_name)
    def execute(self, sql_statement: str, sql_parameters: tuple | list = None, multi: bool = False) -> None:
        # TODO: validate_select_table_name(table_name)
        if sql_parameters:
            quoted_parameters = tuple(
                "'" + str(param) + "'" for param in sql_parameters)
            formatted_sql = sql_statement % quoted_parameters
            sql_parameters_str = ", ".join(quoted_parameters)
        else:
            formatted_sql = sql_statement
            sql_parameters_str = "None"

        formatted_sql = ' '.join(formatted_sql.split())

        self.logger.info(object={
            "formatted_sql": formatted_sql,
            "sql_parameters": sql_parameters_str,
            "sql_statement": sql_statement.replace("\n", " ")
        })
        if MYSQL_VERSION is not None and version_tuple(MYSQL_VERSION) < version_tuple("9.2.0"):
            self.cursor.execute(sql_statement, sql_parameters, multi=multi)
        else:
            self.cursor.execute(sql_statement, sql_parameters)

    def executemany(self, sql_statement: str, sql_parameters: tuple | list = None) -> None:
        try:
            if sql_parameters:  # sql_parameters is list of tuples, each tuple is a row
                sql_parameters_str = [
                    tuple(f'"{param}"' if not isinstance(param, str) else param for param in sql_parameter) for
                    sql_parameter in sql_parameters]
                # Num of placeholders is the same as the num of columns in the table,
                # but there are multiple rows in sql_parameters, so we should add more placeholders before formatting
                sql, values = sql_statement.split("VALUES")
                values = values.strip()
                if not values.startswith("("):
                    values = "(" + values
                if not values.endswith(")"):
                    values = values + ")"
                placeholders = "[" + \
                    ", ".join([values] * len(sql_parameters)) + "]"
                concat_params_tuple = tuple(
                    [param for tup in sql_parameters_str for param in tup])
                formatted_sql = f"{sql} VALUES {placeholders}" % concat_params_tuple
            else:
                formatted_sql = sql_statement
                sql_parameters_str = "None"
            self.logger.info(object={
                "formatted_sql": formatted_sql.replace("\n", " "),
                "sql_parameters": sql_parameters_str,
                "sql_statement": sql_statement.replace("\n", " ")
            })
        except Exception as e:
            self.logger.warning('Unable to format parameters', object={
                "sql_statement": sql_statement,
                "sql_parameters": sql_parameters,
                "error": str(e)
            })

        self.cursor.executemany(sql_statement, sql_parameters)

    def fetchall(self) -> Any:
        result = self.cursor.fetchall()
        return result

    def fetchmany(self, size: int) -> Any:
        result = self.cursor.fetchmany(size)
        return result

    def fetchone(self) -> Any:
        result = self.cursor.fetchone()
        return result

    def description(self) -> list[tuple]:
        """Returns description of columns in a result

        This property returns a list of tuples describing the columns
        in a result set. A tuple is described as follows::

                (column_name,
                 type,
                 None,
                 None,
                 None,
                 None,
                 null_ok,
                 column_flags)  # Addition to PEP-249 specs

        Returns a list of tuples.
        """
        result = self.cursor.description
        return result

    def column_names(self) -> tuple:
        # TODO Please change all result variables to meaningful names i.e. column_names
        result = self.cursor.column_names
        return result

    def lastrowid(self) -> int | None:
        """Returns the value generated for an AUTO_INCREMENT column by the previous INSERT or UPDATE statement."""
        result = self.cursor.lastrowid
        return result

    def get_affected_row_count(self) -> int:
        """Returns the number of rows produced or affected"""
        result = self.cursor.rowcount
        return result

    def get_last_executed_statement(self) -> str:
        result = self.cursor.statement
        return result

    def close(self) -> None:
        if self.__is_closed:
            self.logger.warning('Cursor is already closed')
        try:
            self.cursor.close()
            self.__is_closed = True
        except Exception as exception:
            self.logger.error('Unable to close cursor',
                              object={"exception": exception})

    def is_closed(self) -> bool:
        result = self.__is_closed
        return result





from __future__ import annotations  # Required for type hinting in class methods

from collections import defaultdict
# import paramiko
# from sshtunnel import SSHTunnelForwarder
from os.path import expanduser
from typing import Optional

import mysql.connector
from logger_local.MetaLogger import MetaLogger

from .constants_src import LOGGER_CONNECTOR_CODE_OBJECT
from .cursor import Cursor
from .utils import (get_sql_hostname, get_sql_password, get_sql_username, get_sql_port)

# get_ssh_hostname, get_ssh_username, get_ssh_port, get_private_key_file_path)


connections_pool = defaultdict(dict)  # TODO: test multiple users

home = expanduser('~')


# TODO Can we add hostname, IPv4, IPv6, sql statement? Can we have the same message format in all cases?
class Connector(metaclass=MetaLogger, object=LOGGER_CONNECTOR_CODE_OBJECT):
    def __repr__(self) -> str:
        # Avoid deep reprs that can recurse via connection/cursor references.
        return (f"Connector(host={self.host!r}, schema={self.schema!r}, "
                f"user={self.user!r}, port={self.port!r})")

    @staticmethod
    def connect(schema_name: str, *, user: str = get_sql_username(), ignore_cache: bool = False) -> Connector:
        if ignore_cache and schema_name in connections_pool[user]:
            connections_pool[user].pop(schema_name)

        if connections_pool[user].get(schema_name):
            # is_connected is too slow & apparently very rarely needed.
            # We may develop our is_connected with cache & timeout.
            connector = None
            if connections_pool[user][schema_name].connection.is_connected():
                connector = connections_pool[user][schema_name]
            else:
                # reconnect
                connections_pool[user][schema_name].connection.reconnect()
            #     # TODO We should develop retry mechanism to support password rotation and small network issues.
            #     if connections_pool[user][schema_name].connection.is_connected():
            #         connector = connections_pool[user][schema_name]
            #     else:
            #         connector = Connector(schema_name)
            #         Connector.logger.warning("Reconnect failed, returning a new connection",
            #                                  object={'old connection': connections_pool[user][schema_name],
            #                                          'new connection': connector})
            #         connections_pool[user][schema_name] = connector

        else:
            connector = Connector(schema_name)

        return connector

    def __init__(self, schema_name: str, *,
                 host: str = get_sql_hostname(),
                 user: str = get_sql_username(),
                 password: str = get_sql_password(),
                 port: str = get_sql_port()) -> None:
        # ssh_host: str = get_ssh_hostname()) -> None:
        self.host = host
        self.schema = schema_name
        self.user = user
        self.password = password
        self.port = port
        # TODO: installing crypto cause serverless to fail. We should add it as extra to setup.py
        # self.ssh_host = ssh_host
        # self.ssh_user = None
        # self.ssh_port = None
        # if self.ssh_host:
        #     ssh_user: str = get_ssh_username()
        #     ssh_port: str = get_ssh_port()
        #     self.ssh_user = ssh_user
        #     self.ssh_port = ssh_port
        #     print("home=" + home)
        #     print(get_private_key_file_path())
        #     self.private_key = paramiko.RSAKey.from_private_key_file(home + get_private_key_file_path())

        # Checking host suffix
        # TODO: move to sdk.get_sql_hostname
        # TODO Allow in (prod1 environment_name or when we have SSH_HOSTNAME) also to connect to localhost or 127.0.0.1
        if not (self.host.endswith("circ.zone") or self.host.endswith("circlez.ai")):
            self.logger.warning(
                f"Your RDS_HOSTNAME={self.host} which is not what is expected")
        self.connection: mysql.connector = None
        self._cursor: Optional[Cursor] = None
        self._connect_to_db()
        connections_pool[user][schema_name] = self

    def reconnect(self) -> None:
        # Called when the connection is lost
        self.connection.reconnect()
        self._cursor = self.cursor(close_previous=True)
        self.set_schema(self.schema)

    def _connect_to_db(self):
        # if not self.ssh_host:
        self.connection = mysql.connector.connect(
            host=self.host,
            user=self.user,
            password=self.password,
            database=self.schema,
            port=self.port
        )
        self.logger.info(f"Connected to the database: {self.database_info()}")
        self._cursor = self.connection.cursor()
        self.set_schema(self.schema)
        # else:
        #     # SSH Tunneling
        #     try:
        #         self.tunnel = SSHTunnelForwarder(
        #             (self.ssh_host, int(self.ssh_port)),
        #             ssh_username=self.ssh_user,
        #             ssh_pkey=self.private_key,
        #             remote_bind_address=(self.host, int(self.port)),
        #             local_bind_address=('127.0.0.1',)
        #         )
        #         self.tunnel.start()
        #         self.logger.info(f"SSH tunnel started successfully: {self.tunnel.local_bind_port}")
        #
        #         self.connection = mysql.connector.connect(
        #             user=self.user,
        #             password=self.password,
        #             host='127.0.0.1',
        #             database=self.schema,
        #             port=self.tunnel.local_bind_port
        #         )
        #         self.logger.info(f"Connected to the database via SSH tunnel: {self.database_info()}")
        #         self._cursor = self.connection.cursor()
        #         self.set_schema(self.schema)
        #     except Exception as exception:
        #         self.logger.exception(f"Failed to connect to the database via SSH tunnel: {exception}")
        #         if self.tunnel:
        #             self.tunnel.stop()
        #         raise exception

    def database_info(self) -> str:
        database_info_str = f"host={self.host}, user={self.user}, schema={self.schema}, port={self.port}"
        return database_info_str

    def close(self) -> None:
        try:
            if self._cursor:
                self._cursor.close()
                self.logger.info(f"Cursor closed successfully for schema: {self.schema}")
        except Exception as exception:
            self.logger.error(object={"exception": exception})
        connections_pool[self.user].pop(self.schema, None)
        if self.connection and self.connection.is_connected():
            self.connection.close()
            self.logger.info("Connection closed successfully.")

    def cursor(self, *, close_previous: bool = True, cache_previous: bool = False, dictionary: bool = None,
               buffered: bool = None, raw: bool = None,
               prepared: bool = None, named_tuple: bool = None, cursor_class=None) -> Cursor:
        # if cache_previous and self._cursor:
        #     cursor_instance = self._cursor
        #     return cursor_instance
        # if close_previous and self._cursor:
        #     self._cursor.close()

        cursor_instance = Cursor(self.connection.cursor(
            dictionary=dictionary, buffered=buffered, raw=raw, prepared=prepared,
            cursor_class=cursor_class))
        # named_tuple=named_tuple, cursor_class=cursor_class))  named_tuple is deprecated from mysql-connector-python 9.3.0

        # self._cursor = cursor_instance
        return cursor_instance

    def commit(self) -> None:
        self.connection.commit()

    def set_schema(self, new_schema: Optional[str]) -> None:
        if not new_schema:
            return
        if self.schema == new_schema:
            return

        if self._cursor and self.connection and self.connection.is_connected():
            use_query = f"USE `{new_schema}`;"
            self._cursor.execute(use_query)
            connections_pool[self.user][new_schema] = self
            connections_pool[self.user].pop(self.schema, None)
            self.schema = new_schema
            self.logger.info(f"Switched to schema: {new_schema}")
        else:
            raise Exception(
                "Connection is not established. The database will be used on the next connect.")

    def rollback(self):
        self.connection.rollback()

    def start_transaction(self):
        self.connection.start_transaction()


